"""Tests for module imports."""

import importlib
import os
import pkgutil
import subprocess
import sys
from collections.abc import Sequence

import pytest
from pytest import param

pytestmark = pytest.mark.skipif(
    os.environ.get("BAYBE_TEST_ENV") != "FULLTEST",
    reason="Only possible in FULLTEST environment.",
)

_EAGER_LOADING_EXIT_CODE = 42


def find_modules() -> list[str]:
    """Return all BayBE module names."""
    package = importlib.import_module("baybe")
    return [
        name
        for _, name, _ in pkgutil.walk_packages(
            path=package.__path__, prefix=package.__name__ + "."
        )
    ]


def make_import_check(modules: Sequence[str], target: str) -> str:
    """Create code that tests if importing the given modules also imports the target.

    Args:
        modules: The modules to be imported by the created code.
        target: The target module whose presence is to be checked after the import.

    Returns:
        Code that signals the presence of the target via a non-zero exit code.
    """
    imports = "\n".join([f"import {module}" for module in modules])
    return "\n".join(
        [
            "import sys",
            f"{imports}",
            f"hit = '{target}' in sys.modules.keys()",
            f"exit({_EAGER_LOADING_EXIT_CODE} if hit else 0)",
        ]
    )


@pytest.mark.parametrize("module", find_modules())
def test_imports(module: str):
    """All modules can be imported without throwing errors."""
    importlib.import_module(module)


WHITELISTS = {
    "torch": [
        "baybe.acquisition.partial",
        "baybe.acquisition._builder",
        "baybe.surrogates._adapter",
        "baybe.utils.torch",
    ],
    "scipy": [
        "baybe._optional.chem",
        "baybe._optional.insights",
        "baybe._optional.ngboost",
        "baybe.acquisition._builder",
        "baybe.acquisition.partial",
        "baybe.insights",
        "baybe.insights.shap",
        "baybe.surrogates._adapter",
        "baybe.utils.chemistry",
        "baybe.utils.clustering_algorithms",
        "baybe.utils.clustering_algorithms.third_party",
        "baybe.utils.clustering_algorithms.third_party.kmedoids",
    ],
    "sklearn": [
        "baybe._optional.chem",
        "baybe._optional.insights",
        "baybe._optional.ngboost",
        "baybe.insights",
        "baybe.insights.shap",
        "baybe.utils.chemistry",
        "baybe.utils.clustering_algorithms",
        "baybe.utils.clustering_algorithms.third_party",
        "baybe.utils.clustering_algorithms.third_party.kmedoids",
    ],
}
"""Modules (dict values) for which certain imports (dict keys) are permitted."""


@pytest.mark.parametrize(
    ("target", "whitelist"), [param(k, v, id=k) for k, v in WHITELISTS.items()]
)
def test_lazy_loading(target: str, whitelist: Sequence[str]):
    """The target does not appear in the module list after loading BayBE modules."""
    all_modules = find_modules()
    assert (w in all_modules for w in whitelist)
    modules = [i for i in all_modules if i not in whitelist]
    code = make_import_check(modules, target)
    python_interpreter = sys.executable
    result = subprocess.call([python_interpreter, "-c", code])
    assert result == 0


@pytest.mark.parametrize(
    ("target", "module"),
    [param(k, m, id=f"{k}-{m}") for k, v in WHITELISTS.items() for m in v],
)
def test_whitelist_modules_are_true_positives(target, module):
    """The whitelisted modules actually import the target."""
    code = make_import_check([module], target)
    python_interpreter = sys.executable
    result = subprocess.call([python_interpreter, "-c", code])
    assert result == _EAGER_LOADING_EXIT_CODE



"""Functionality for building BoTorch acquisition functions."""

from collections.abc import Callable, Iterable
from functools import cached_property
from inspect import signature
from types import MappingProxyType
from typing import Any

import botorch.acquisition as bo_acqf
import pandas as pd
import torch
from attrs import asdict, define, field, fields
from attrs.validators import instance_of, optional
from botorch.acquisition import AcquisitionFunction as BoAcquisitionFunction
from botorch.acquisition.monte_carlo import MCAcquisitionObjective as BoObjective
from botorch.acquisition.multi_objective import WeightedMCMultiOutputObjective
from botorch.acquisition.objective import LinearMCObjective
from botorch.models.model import Model
from torch import Tensor

from baybe.acquisition.acqfs import (
    PosteriorStandardDeviation,
    _ExpectedHypervolumeImprovement,
    qNegIntegratedPosteriorVariance,
    qPosteriorStandardDeviation,
    qThompsonSampling,
)
from baybe.acquisition.base import AcquisitionFunction, _get_botorch_acqf_class
from baybe.exceptions import IncompleteMeasurementsError
from baybe.objectives.base import Objective
from baybe.objectives.desirability import DesirabilityObjective
from baybe.objectives.pareto import ParetoObjective
from baybe.objectives.single import SingleTargetObjective
from baybe.searchspace.core import SearchSpace
from baybe.surrogates.base import SurrogateProtocol
from baybe.targets.enum import TargetMode
from baybe.targets.numerical import NumericalTarget
from baybe.utils.basic import is_all_instance, match_attributes
from baybe.utils.dataframe import handle_missing_values, to_tensor


def opt_v(x: Any, /) -> Callable:
    """Shorthand for an optional attrs isinstance validator."""  # noqa: D401
    return optional(instance_of(x))


@define(kw_only=True)
class BotorchAcquisitionArgs:
    """The collection of (possible) arguments for BoTorch acquisition functions."""

    # Always required
    model: Model = field(validator=instance_of(Model))

    # Optional, depending on the specific acquisition function being used
    best_f: float | None = field(default=None, validator=opt_v(float))
    beta: float | None = field(default=None, validator=opt_v(float))
    maximize: bool | None = field(default=None, validator=opt_v(bool))
    mc_points: Tensor | None = field(default=None, validator=opt_v(Tensor))
    num_fantasies: int | None = field(default=None, validator=opt_v(int))
    objective: BoObjective | None = field(default=None, validator=opt_v(BoObjective))
    prune_baseline: bool | None = field(default=None, validator=opt_v(bool))
    ref_point: Tensor | None = field(default=None, validator=opt_v(Tensor))
    X_baseline: Tensor | None = field(default=None, validator=opt_v(Tensor))
    X_pending: Tensor | None = field(default=None, validator=opt_v(Tensor))

    def collect(self) -> dict[str, Any]:
        """Collect the assigned arguments into a dictionary."""
        return asdict(self, filter=lambda _, x: x is not None)


flds = fields(BotorchAcquisitionArgs)
"""Shorthand for the argument field references."""


@define
class BotorchAcquisitionFunctionBuilder:
    """A class for building BoTorch acquisition functions from BayBE objects."""

    # The BayBE acquisition function to be translated
    acqf: AcquisitionFunction = field(validator=instance_of(AcquisitionFunction))

    # The (pre-validated) BayBE objects
    surrogate: SurrogateProtocol = field()
    searchspace: SearchSpace = field()
    objective: Objective = field()
    measurements: pd.DataFrame = field()
    pending_experiments: pd.DataFrame | None = field(default=None)

    # Context shared across building methods
    _args: BotorchAcquisitionArgs = field(init=False)
    _botorch_acqf_cls: BoAcquisitionFunction = field(init=False)
    _signature: MappingProxyType = field(init=False)
    _set_best_f_called: bool = field(init=False, default=False)

    def __attrs_post_init__(self) -> None:
        """Initialize the building process."""
        # Retrieve botorch acquisition function class and match attributes
        self._botorch_acqf_cls = _get_botorch_acqf_class(type(self.acqf))
        self._signature = signature(self._botorch_acqf_cls).parameters
        args, _ = match_attributes(
            self.acqf,
            self._botorch_acqf_cls.__init__,
            ignore=self.acqf._non_botorch_attrs,
        )

        # Pre-populate the acqf arguments with the content of the BayBE acqf
        self._args = BotorchAcquisitionArgs(model=self.surrogate.to_botorch(), **args)

    @cached_property
    def _botorch_surrogate(self) -> Model:
        """The botorch surrogate object."""
        return self.surrogate.to_botorch()

    @property
    def _maximize_flags(self) -> list[bool]:
        """Booleans indicating which target is to be minimized/maximized."""
        assert is_all_instance(self.objective.targets, NumericalTarget)
        return [t.mode is not TargetMode.MIN for t in self.objective.targets]

    @property
    def _multiplier(self) -> list[float]:
        """Signs indicating which target is to be minimized/maximized."""
        return [1.0 if m else -1.0 for m in self._maximize_flags]

    @cached_property
    def _train_x(self) -> pd.DataFrame:
        """The training parameter values."""
        return self.searchspace.transform(self.measurements, allow_extra=True)

    @cached_property
    def _target_configurations(self) -> pd.DataFrame:
        """The target configurations used for reference point calculation.

        Only completely measured points are considered.

        Returns:
            A dataframe of target configurations.

        Raises:
            ValueError: If no complete measurement exists.
        """
        configurations = handle_missing_values(
            self.measurements[[t.name for t in self.objective.targets]],
            [t.name for t in self.objective.targets],
            drop=True,
        )

        # TODO: A smarter treatment might be possible in the case that not at least
        #  one complete measurement exists, e.g. by considering target bounds or other
        #  heuristics.
        if configurations.empty:
            raise IncompleteMeasurementsError(
                f"For calculating a default reference point, at least one "
                f"configuration must have a measured value for all targets. You can "
                f"fix this by setting the "
                f"'{fields(_ExpectedHypervolumeImprovement).reference_point.name}' "
                f"argument of '{self.acqf.__class__.__name__}' explicitly."
            )

        return self.objective.transform(configurations)

    def build(self) -> BoAcquisitionFunction:
        """Build the BoTorch acquisition function object."""
        # Set context-specific parameters
        self._set_best_f()
        self._invert_optimization_direction()
        self._set_X_baseline()
        self._set_X_pending()
        self._set_mc_points()
        self._set_ref_point()

        botorch_acqf = self._botorch_acqf_cls(**self._args.collect())
        self.set_default_sample_shape(botorch_acqf)

        return botorch_acqf

    def _invert_optimization_direction(self) -> None:
        """Invert optimization direction for minimization targets."""
        # ``best_f`` must have been already set (for the inversion below to work)
        assert self._set_best_f_called

        if issubclass(
            type(self.acqf),
            (
                qNegIntegratedPosteriorVariance,
                PosteriorStandardDeviation,
                qPosteriorStandardDeviation,
            ),
        ):
            # No action needed for the active learning acquisition functions:
            # - PSTD: Minimization happens by setting `maximize=False`, which is
            #   already take care of by auto-matching attributes
            # - qPSTD and qNIPV do not support minimization yet
            # In both cases, the setting is independent of the target mode.
            return

        match self.objective:
            case SingleTargetObjective(NumericalTarget(mode=TargetMode.MIN)):
                if issubclass(self._botorch_acqf_cls, bo_acqf.MCAcquisitionFunction):
                    if self._args.best_f is not None:
                        self._args.best_f *= -1.0
                    self._args.objective = LinearMCObjective(torch.tensor([-1.0]))
                elif issubclass(
                    self._botorch_acqf_cls, bo_acqf.AnalyticAcquisitionFunction
                ):
                    self._args.maximize = False

            case ParetoObjective():
                self._args.objective = WeightedMCMultiOutputObjective(
                    torch.tensor(self._multiplier)
                )

    def _set_best_f(self) -> None:
        """Set BoTorch's ``best_f`` argument."""
        self._set_best_f_called = True

        if flds.best_f.name not in self._signature:
            return

        post_mean = self._botorch_surrogate.posterior(to_tensor(self._train_x)).mean

        match self.objective:
            case SingleTargetObjective(NumericalTarget(mode=TargetMode.MIN)):
                self._args.best_f = post_mean.min().item()
            case SingleTargetObjective() | DesirabilityObjective():
                self._args.best_f = post_mean.max().item()

    def set_default_sample_shape(self, acqf: BoAcquisitionFunction, /):
        """Apply temporary workaround for Thompson sampling."""
        if not isinstance(self.acqf, qThompsonSampling):
            return

        assert hasattr(acqf, "_default_sample_shape")
        acqf._default_sample_shape = torch.Size([self.acqf.n_mc_samples])

    def _set_mc_points(self) -> None:
        """Set BoTorch's ``mc_points`` argument."""
        if flds.mc_points.name not in self._signature:
            return

        assert isinstance(self.acqf, qNegIntegratedPosteriorVariance)
        self._args.mc_points = to_tensor(
            self.acqf.get_integration_points(self.searchspace)
        )

    def _set_ref_point(self) -> None:
        """Set BoTorch's ``ref_point`` argument."""
        if flds.ref_point.name not in self._signature:
            return

        assert isinstance(self.acqf, _ExpectedHypervolumeImprovement)

        if isinstance(ref_point := self.acqf.reference_point, Iterable):
            self._args.ref_point = torch.tensor(
                [p * m for p, m in zip(ref_point, self._multiplier, strict=True)]
            )
        else:
            kwargs = {} if ref_point is None else {"factor": ref_point}
            self._args.ref_point = torch.tensor(
                self.acqf.compute_ref_point(
                    self._target_configurations.to_numpy(),
                    self._maximize_flags,
                    **kwargs,
                )
                * self._multiplier
            )

    def _set_X_baseline(self) -> None:
        """Set BoTorch's ``X_baseline`` argument."""
        if flds.X_baseline.name not in self._signature:
            return

        self._args.X_baseline = to_tensor(self._train_x)

    def _set_X_pending(self) -> None:
        """Set BoTorch's ``X_pending`` argument."""
        if self.pending_experiments is None:
            return

        pending_x = self.searchspace.transform(
            self.pending_experiments, allow_extra=True
        )
        self._args.X_pending = to_tensor(pending_x)



"""Parameter utilities."""

from collections.abc import Callable, Collection
from functools import partial
from typing import Any, TypeVar

import numpy as np
import numpy.typing as npt
import pandas as pd
from attrs import evolve

from baybe.parameters.base import Parameter
from baybe.parameters.numerical import (
    NumericalContinuousParameter,
)
from baybe.utils.interval import Interval

_TParameter = TypeVar("_TParameter", bound=Parameter)


def get_parameters_from_dataframe(
    df: pd.DataFrame,
    factory: Callable[[str, Collection[Any]], _TParameter],
    parameters: Collection[_TParameter] | None = None,
) -> list[_TParameter]:
    """Create a list of parameters from a dataframe.

    Returns one parameter for each column of the given dataframe. By default,
    the parameters are created using the provided factory, which takes the name
    of the column and its unique values as arguments. However, there is also
    the possibility to provide explicit parameter objects with names matching specific
    columns of the dataframe, to bypass the parameter factory creation for those
    columns. This allows finer control, for example, to specify custom parameter
    attributes (e.g. specific optional arguments) compared to what would be provided
    by the factory. Still, the pre-specified parameters are validated to ensure that
    they are compatible with the contents of the dataframe.

    Args:
        df: The dataframe from which to create the parameters.
        factory: A callable creating parameter objects for the columns from the column
            name and the unique column values.
        parameters: An optional list of parameter objects to bypass the factory
            creation for columns whose names match with the parameter names.

    Returns:
        The combined parameter list, containing both the (validated) pre-specified
        parameters and the parameters inferred from the dataframe.

    Raises:
        ValueError: If several parameters with identical names are provided.
        ValueError: If a parameter was specified for which no match was found.
    """
    # Turn the pre-specified parameters into a dict and check for duplicate names
    specified_params: dict[str, _TParameter] = {}
    if parameters is not None:
        for param in parameters:
            if param.name in specified_params:
                raise ValueError(
                    f"You provided several parameters with the name '{param.name}'."
                )
            specified_params[param.name] = param

    # Try to find a parameter match for each dataframe column
    parameters = []
    for name, series in df.items():
        assert isinstance(name, str), (
            "The given dataframe must only contain string-valued column names."
        )
        unique_values = series.unique()

        # If a match is found, assert that the values are in range
        if match := specified_params.pop(name, None):
            if not all(match.is_in_range(x) for x in unique_values):
                raise ValueError(
                    f"The dataframe column '{name}' contains the values "
                    f"{unique_values}, which are outside the range of {match}."
                )
            parameters.append(match)

        # Otherwise, create a new parameter using the factory
        else:
            param = factory(name, unique_values)
            parameters.append(param)

    # By now, all pre-specified parameters must have been used
    if specified_params:
        raise ValueError(
            f"For the parameter(s) {list(specified_params.keys())}, "
            f"no match could be found in the given dataframe."
        )

    return parameters


def sort_parameters(parameters: Collection[Parameter]) -> tuple[Parameter, ...]:
    """Sort parameters alphabetically by their names."""
    return tuple(sorted(parameters, key=lambda p: p.name))


def activate_parameter(
    parameter: NumericalContinuousParameter, thresholds: Interval
) -> NumericalContinuousParameter:
    """Force-activates a given parameter by moving its bounds away from zero.

    A parameter is considered active if its value falls outside the specified threshold
    interval. Force-activating a parameter adjusts its range to ensure it cannot take
    values within this interval. Parameters that are inherently active, due to their
    original value ranges not overlapping with the inactivity interval, remain
    unchanged.

    Important:
        A parameter whose range includes zero but extends beyond the threshold interval
        on both sides remains unchanged, because the corresponding activated parameter
        would no longer have a continuous value range.

    Args:
        parameter: The parameter to be activated.
        thresholds: The considered parameter (in)activity thresholds.

    Returns:
        A copy of the parameter with adjusted bounds.

    Raises:
        ValueError: If the threshold interval does not contain zero.
        ValueError: If the parameter cannot be activated since both its bounds are
            in the inactive range.
        NotImplementedError: In situations that cannot be encountered in a regular
            recommendation context since prevented by other validation measures (for
            example, edge case of an activation that would result in a single active
            point when the open inactive interval aligns on one side with the parameter
            bounds). For such situations, the behavior of the function is not defined.
    """

    def is_fraction(value: float, reference: float) -> bool:
        """Check if the given value is a fraction of a specified reference value."""
        if value == 0.0:
            return reference == 0.0
        return reference / value > 1.0

    lower_bound = parameter.bounds.lower
    upper_bound = parameter.bounds.upper

    if not thresholds.contains(0.0):
        raise ValueError(
            f"The thresholds must cover zero but ({thresholds.lower}, "
            f"{thresholds.upper}) was given."
        )

    if not (
        is_fraction(thresholds.lower, lower_bound)
        and is_fraction(thresholds.upper, upper_bound)
    ):
        raise NotImplementedError(
            "This function is implemented only for the case when "
            "thresholds is a proper sub-interval of the parameter bounds."
        )

    # Callable checking whether the argument is within the inactive range
    _is_inactive = partial(
        is_inactive,
        lower_threshold=thresholds.lower,
        upper_threshold=thresholds.upper,
    )

    # When the upper bound is in inactive range, move it to the lower threshold of the
    # inactive region
    if not _is_inactive(lower_bound) and _is_inactive(upper_bound):
        return evolve(parameter, bounds=(lower_bound, thresholds.lower))

    # When the lower bound is in inactive range, move it to the upper threshold of
    # the inactive region
    if not _is_inactive(upper_bound) and _is_inactive(lower_bound):
        return evolve(parameter, bounds=(thresholds.upper, upper_bound))

    # When the parameter is already trivially active (or activating it would tear
    # its value range apart)
    return parameter


def is_inactive(
    x: npt.ArrayLike, /, lower_threshold: npt.ArrayLike, upper_threshold: npt.ArrayLike
) -> np.ndarray:
    """Check if the given values are inactive (i.e. can be treated as zero).

    A value is considered inactive when at least one of the following is true:
    * The value lies in the open interval specified by the given thresholds.
    * The value is zero.

    Args:
        x: An array-like object containing numeric values.
        lower_threshold: The (broadcastable) lower thresholds of the inactive regions.
        upper_threshold: The (broadcastable) upper thresholds of the inactive regions.

    Returns:
        A Boolean-valued numpy array indicating which elements are inactive.
    """
    x = np.asarray(x)
    lower_threshold = np.asarray(lower_threshold)
    upper_threshold = np.asarray(upper_threshold)
    return ((x > lower_threshold) & (x < upper_threshold)) | (x == 0.0)



"""Optional xyzpy import."""

from baybe.exceptions import OptionalImportError

try:
    import xyzpy
except ModuleNotFoundError as ex:
    raise OptionalImportError(name="xyzpy", group="simulation") from ex

__all__ = [
    "xyzpy",
]



"""Tests for the substance parameter."""

import pytest

from baybe._optional.info import CHEM_INSTALLED
from baybe.parameters.enum import SubstanceEncoding

from .conftest import run_iterations


@pytest.mark.skipif(
    not CHEM_INSTALLED, reason="Optional chem dependency not installed."
)
@pytest.mark.parametrize(
    "parameter_names",
    [["Categorical_1", f"Substance_1_{enc.name}"] for enc in SubstanceEncoding],
    ids=[enc.name for enc in SubstanceEncoding],
)
@pytest.mark.parametrize("n_grid_points", [8], ids=["g8"])
@pytest.mark.parametrize("batch_size", [1], ids=["b3"])
def test_run_iterations(campaign, batch_size, n_iterations):
    """Test running some iterations with fake results and a substance parameter."""
    run_iterations(campaign, n_iterations, batch_size)


@pytest.mark.skipif(
    not CHEM_INSTALLED, reason="Optional chem dependency not installed."
)
def test_degenerate_comp_df():
    """Test a degenerate comp_df is detected and fixed numerically."""
    from baybe.parameters import SubstanceParameter

    # These molecules are know to cause a degenerate representation with rdkit fp's
    dict_base = {
        "Potassium acetate": r"O=C([O-])C.[K+]",
        "Potassium pivalate": r"O=C([O-])C(C)(C)C.[K+]",
        "Cesium acetate": r"O=C([O-])C.[Cs+]",
        "Cesium pivalate": r"O=C([O-])C(C)(C)C.[Cs+]",
    }
    p = SubstanceParameter(name="p", data=dict_base, encoding="RDKITFINGERPRINT")

    assert not p.comp_df.duplicated().any(), (
        "A degenerate comp_df was not correctly treated."
    )



"""Base classes for all acquisition functions."""

from __future__ import annotations

import gc
from abc import ABC
from typing import TYPE_CHECKING, ClassVar, Literal, overload

import pandas as pd
from attrs import define

from baybe.exceptions import (
    IncompatibleAcquisitionFunctionError,
    UnidentifiedSubclassError,
)
from baybe.objectives.base import Objective
from baybe.searchspace.core import SearchSpace
from baybe.serialization.core import (
    converter,
    get_base_structure_hook,
    unstructure_base,
)
from baybe.serialization.mixin import SerialMixin
from baybe.surrogates.base import SurrogateProtocol
from baybe.utils.basic import classproperty
from baybe.utils.boolean import is_abstract
from baybe.utils.dataframe import to_tensor

if TYPE_CHECKING:
    from botorch.acquisition import AcquisitionFunction as BotorchAcquisitionFunction


@define(frozen=True)
class AcquisitionFunction(ABC, SerialMixin):
    """Abstract base class for all acquisition functions."""

    abbreviation: ClassVar[str]
    """An alternative name for type resolution."""

    supports_multi_output: ClassVar[bool] = False
    """Whether this acquisition function can handle models with multiple outputs."""

    @classproperty
    def supports_batching(cls) -> bool:
        """Flag indicating whether batch recommendation is supported."""
        return cls.abbreviation.startswith("q")

    @classproperty
    def supports_pending_experiments(cls) -> bool:
        """Flag indicating whether pending experiments are supported.

        This is based on the same mechanism underlying batched recommendations.
        """
        return cls.supports_batching

    @classproperty
    def _non_botorch_attrs(cls) -> tuple[str, ...]:
        """Names of attributes that are not passed to the BoTorch constructor."""
        return ()

    def to_botorch(
        self,
        surrogate: SurrogateProtocol,
        searchspace: SearchSpace,
        objective: Objective,
        measurements: pd.DataFrame,
        pending_experiments: pd.DataFrame | None = None,
    ) -> BotorchAcquisitionFunction:
        """Create the botorch-ready representation of the function.

        The required structure of `measurements` is specified in
        :meth:`baybe.recommenders.base.RecommenderProtocol.recommend`.
        """
        from baybe.acquisition._builder import BotorchAcquisitionFunctionBuilder

        if pending_experiments is not None and not self.supports_pending_experiments:
            raise IncompatibleAcquisitionFunctionError(
                f"The chosen acquisition function of type '{self.__class__.__name__}' "
                f"does not support pending experiments."
            )

        return BotorchAcquisitionFunctionBuilder(
            self, surrogate, searchspace, objective, measurements, pending_experiments
        ).build()

    @overload
    def evaluate(
        self,
        candidates: pd.DataFrame,
        surrogate: SurrogateProtocol,
        searchspace: SearchSpace,
        objective: Objective,
        measurements: pd.DataFrame,
        pending_experiments: pd.DataFrame | None = None,
        *,
        jointly: Literal[True],
    ) -> float: ...

    @overload
    def evaluate(
        self,
        candidates: pd.DataFrame,
        surrogate: SurrogateProtocol,
        searchspace: SearchSpace,
        objective: Objective,
        measurements: pd.DataFrame,
        pending_experiments: pd.DataFrame | None = None,
        *,
        jointly: Literal[False] = False,
    ) -> pd.Series: ...

    def evaluate(
        self,
        candidates: pd.DataFrame,
        surrogate: SurrogateProtocol,
        searchspace: SearchSpace,
        objective: Objective,
        measurements: pd.DataFrame,
        pending_experiments: pd.DataFrame | None = None,
        *,
        jointly: bool = False,
    ) -> pd.Series | float:
        """Get the acquisition values for the given candidates.

        Args:
            candidates: The candidate points in experimental representation.
                For details, see :meth:`baybe.surrogates.base.Surrogate.posterior`.
            surrogate: The surrogate model to use for the acquisition function.
            searchspace: The search space.
                See :meth:`baybe.recommenders.base.RecommenderProtocol.recommend`.
            objective: The objective.
                See :meth:`baybe.recommenders.base.RecommenderProtocol.recommend`.
            measurements: Available experimentation data.
                See :meth:`baybe.recommenders.base.RecommenderProtocol.recommend`.
            pending_experiments: Optional pending experiments.
                See :meth:`baybe.recommenders.base.RecommenderProtocol.recommend`.
            jointly: If ``False``, the acquisition values are computed independently
                for each candidate. If ``True``, a single joint acquisition value is
                computed for the entire candidate set.

        Returns:
            Depending on the joint mode, either a single batch acquisition value or a
            series of individual acquisition values.
        """
        import torch

        # Assemble the Botorch acquisition function and its input
        botorch_acqf = self.to_botorch(
            surrogate, searchspace, objective, measurements, pending_experiments
        )
        comp = to_tensor(searchspace.transform(candidates))

        # Depending on joint mode, evaluate using t- or q-batching
        in_ = comp if jointly else comp.unsqueeze(-2)
        with torch.no_grad():
            out = botorch_acqf(in_)
        if jointly:
            return out.item()
        return pd.Series(out.numpy(), index=candidates.index)


def _get_botorch_acqf_class(
    baybe_acqf_cls: type[AcquisitionFunction], /
) -> type[BotorchAcquisitionFunction]:
    """Extract the BoTorch acquisition class for the given BayBE acquisition class."""
    import botorch

    for cls in baybe_acqf_cls.mro():
        if (
            acqf_cls := getattr(botorch.acquisition, cls.__name__, False)
            or getattr(botorch.acquisition.multi_objective, cls.__name__, False)
            or getattr(botorch.acquisition.multi_objective.parego, cls.__name__, False)
        ):
            if is_abstract(acqf_cls):
                continue
            return acqf_cls  # type: ignore

    raise UnidentifiedSubclassError(
        f"No BoTorch acquisition function class match found for "
        f"'{baybe_acqf_cls.__name__}'."
    )


# Register (un-)structure hooks
converter.register_structure_hook(
    AcquisitionFunction, get_base_structure_hook(AcquisitionFunction)
)
converter.register_unstructure_hook(AcquisitionFunction, unstructure_base)

# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Utilities related to data augmentation."""

from collections.abc import Collection, Sequence
from itertools import permutations, product

import pandas as pd


def df_apply_permutation_augmentation(
    df: pd.DataFrame,
    column_groups: Sequence[Sequence[str]],
) -> pd.DataFrame:
    """Augment a dataframe if permutation invariant columns are present.

    Args:
        df: The dataframe that should be augmented.
        column_groups: Sequences of permutation invariant columns. The n'th column in
            each group will be permuted together with each n'th column in the other
            groups.

    Returns:
        The augmented dataframe containing the original one. Augmented row indices are
        identical with the index of their original row.

    Raises:
        ValueError: If less than two column groups are given.
        ValueError: If a column group is empty.
        ValueError: If the column groups have differing amounts of entries.

    Examples:
        >>> df = pd.DataFrame({'A1':[1,2],'A2':[3,4], 'B1': [5, 6], 'B2': [7, 8]})
        >>> df
           A1  A2  B1  B2
        0   1   3   5   7
        1   2   4   6   8

        >>> column_groups = [['A1'], ['A2']]
        >>> dfa = df_apply_permutation_augmentation(df, column_groups)
        >>> dfa
           A1  A2  B1  B2
        0   1   3   5   7
        0   3   1   5   7
        1   2   4   6   8
        1   4   2   6   8

        >>> column_groups = [['A1', 'B1'], ['A2', 'B2']]
        >>> dfa = df_apply_permutation_augmentation(df, column_groups)
        >>> dfa
           A1  A2  B1  B2
        0   1   3   5   7
        0   3   1   7   5
        1   2   4   6   8
        1   4   2   8   6
    """
    # Validation
    if len(column_groups) < 2:
        raise ValueError(
            "When augmenting permutation invariance, at least two column sequences "
            "must be given."
        )

    if len({len(seq) for seq in column_groups}) != 1:
        raise ValueError(
            "Permutation augmentation can only work if the amount of columns in each "
            "sequence is the same."
        )
    elif len(column_groups[0]) < 1:
        raise ValueError(
            "Permutation augmentation can only work if each column group has at "
            "least one entry."
        )

    # Augmentation Loop
    new_rows: list[pd.DataFrame] = []
    idx_permutation = list(permutations(range(len(column_groups))))
    for _, row in df.iterrows():
        # For each row in the original df, collect all its permutations
        to_add = []
        for perm in idx_permutation:
            new_row = row.copy()

            # Permute columns, this is done separately for each tuple of columns that
            # belong together
            for deps in map(list, zip(*column_groups)):
                new_row[deps] = row[[deps[k] for k in perm]]

            to_add.append(new_row)

        # Store augmented rows, keeping the index of their original row
        new_rows.append(
            pd.DataFrame(to_add, columns=df.columns, index=[row.name] * len(to_add))
        )

    return pd.concat(new_rows)


def df_apply_dependency_augmentation(
    df: pd.DataFrame,
    causing: tuple[str, Sequence],
    affected: Collection[tuple[str, Sequence]],
) -> pd.DataFrame:
    """Augment a dataframe if dependency invariant columns are present.

    This works with the concept of column-values pairs for causing and affected column.
    Any row present where the specified causing column has one of the provided values
    will trigger an augmentation on the affected columns. The latter are augmented by
    going through all their invariant values and adding respective new rows.

    Args:
        df: The dataframe that should be augmented.
        causing: Causing column name and its causing values.
        affected: Affected columns and their invariant values.

    Returns:
        The augmented dataframe containing the original one. Augmented row indices are
        identical with the index of their original row.

    Examples:
        >>> df = pd.DataFrame({'A':[0,1],'B':[2,3], 'C': [5, 5], 'D': [6, 7]})
        >>> df
           A  B  C  D
        0  0  2  5  6
        1  1  3  5  7

        >>> causing = ('A', [0])
        >>> affected = [('B', [2,3,4])]
        >>> dfa = df_apply_dependency_augmentation(df, causing, affected)
        >>> dfa
           A  B  C  D
        0  0  2  5  6
        0  0  3  5  6
        0  0  4  5  6
        1  1  3  5  7

        >>> causing = ('A', [0])
        >>> affected = [('B', [2,3,4])]
        >>> dfa = df_apply_dependency_augmentation(df, causing, affected)
        >>> dfa
           A  B  C  D
        0  0  2  5  6
        0  0  3  5  6
        0  0  4  5  6
        1  1  3  5  7

        >>> causing = ('A', [0, 1])
        >>> affected = [('B', [2,3])]
        >>> dfa = df_apply_dependency_augmentation(df, causing, affected)
        >>> dfa
           A  B  C  D
        0  0  2  5  6
        0  0  3  5  6
        1  1  2  5  7
        1  1  3  5  7

        >>> causing = ('A', [0])
        >>> affected = [('B', [2,3]), ('C', [5, 6])]
        >>> dfa = df_apply_dependency_augmentation(df, causing, affected)
        >>> dfa
           A  B  C  D
        0  0  2  5  6
        0  0  2  6  6
        0  0  3  5  6
        0  0  3  6  6
        1  1  3  5  7
    """
    new_rows: list[pd.DataFrame] = []
    col_causing, vals_causing = causing
    affected_cols, affected_inv_vals = zip(*affected)
    affected_inv_vals_combinations = list(product(*affected_inv_vals))

    # Iterate through all rows that have a causing value in the respective column.
    for _, row in df.iterrows():
        to_add = []
        if row[col_causing] not in vals_causing:
            # Just keep unaffected rows without augmentation
            to_add.append(row)
        else:
            # Create augmented rows by assigning the affected columns all possible
            # values
            to_add += [
                pd.Series({**row.to_dict(), **dict(zip(affected_cols, values))})
                for values in affected_inv_vals_combinations
            ]

        # Store augmented rows, keeping the index of their original row
        new_rows.append(
            pd.DataFrame(to_add, columns=df.columns, index=[row.name] * len(to_add))
        )

    return pd.concat(new_rows)



"""Tests pending experiments mechanism."""

import warnings

import numpy as np
import pandas as pd
import pytest
from pytest import param

from baybe.acquisition.base import AcquisitionFunction
from baybe.exceptions import IncompatibleAcquisitionFunctionError, UnusedObjectWarning
from baybe.recommenders import (
    BotorchRecommender,
    FPSRecommender,
    GaussianMixtureClusteringRecommender,
    KMeansClusteringRecommender,
    NaiveHybridSpaceRecommender,
    PAMClusteringRecommender,
    TwoPhaseMetaRecommender,
)
from baybe.searchspace.core import SearchSpaceType
from baybe.utils.basic import get_subclasses
from baybe.utils.dataframe import add_parameter_noise
from baybe.utils.random import temporary_seed

_discrete_params = ["Categorical_1", "Switch_1", "Num_disc_1"]
_continuous_params = ["Conti_finite1", "Conti_finite2", "Conti_finite3"]
_hybrid_params = ["Categorical_1", "Num_disc_1", "Conti_finite1", "Conti_finite2"]


@pytest.mark.parametrize(
    "parameter_names, recommender",
    [
        param(
            _discrete_params,
            FPSRecommender(),
            id="fps_discrete",
        ),
        param(
            _discrete_params,
            PAMClusteringRecommender(),
            id="pam_discrete",
        ),
        param(
            _discrete_params,
            KMeansClusteringRecommender(),
            id="kmeans_discrete",
        ),
        param(
            _discrete_params,
            GaussianMixtureClusteringRecommender(),
            id="gm_discrete",
        ),
        param(
            _discrete_params,
            TwoPhaseMetaRecommender(recommender=BotorchRecommender()),
            id="botorch_discrete",
        ),
        param(
            _continuous_params,
            TwoPhaseMetaRecommender(recommender=BotorchRecommender()),
            id="botorch_continuous",
        ),
        param(
            _hybrid_params,
            TwoPhaseMetaRecommender(recommender=BotorchRecommender()),
            id="botorch_hybrid",
        ),
        param(
            _discrete_params,
            TwoPhaseMetaRecommender(recommender=BotorchRecommender()),
            id="botorch_discrete_allow",
        ),
        param(
            _continuous_params,
            TwoPhaseMetaRecommender(recommender=BotorchRecommender()),
            id="botorch_continuous_allow",
        ),
        param(
            _hybrid_params,
            TwoPhaseMetaRecommender(recommender=BotorchRecommender()),
            id="botorch_hybrid_allow",
        ),
        param(
            _discrete_params,
            NaiveHybridSpaceRecommender(
                disc_recommender=FPSRecommender(),
            ),
            id="naive1_discrete",
        ),
        param(
            _discrete_params,
            NaiveHybridSpaceRecommender(
                disc_recommender=KMeansClusteringRecommender(),
            ),
            id="naive2_discrete",
        ),
    ],
)
@pytest.mark.parametrize("n_grid_points", [8], ids=["grid8"])
def test_pending_points(campaign, batch_size, fake_measurements):
    """Test there is no recommendation overlap if pending experiments are specified."""
    warnings.filterwarnings("ignore", category=UnusedObjectWarning)

    # Repeated recommendations explicitly need to be allowed or the potential overlap
    # will be avoided trivially
    if campaign.searchspace.type == SearchSpaceType.DISCRETE:
        campaign.allow_recommending_already_recommended = True
        campaign.allow_recommending_already_measured = True

    # Add some initial measurements
    campaign.add_measurements(fake_measurements)

    # Get recommendations and set them as pending experiments while getting another set
    # Fix the random seed for each recommend call to limit influence of randomness in
    # some recommenders which could also trivially avoid overlap
    with temporary_seed(1337):
        rec1 = campaign.recommend(batch_size)
    campaign._cached_recommendation = pd.DataFrame()  # ensure no recommendation cache
    with temporary_seed(1337):
        rec2 = campaign.recommend(batch_size=batch_size, pending_experiments=rec1)

    # Assert they have no overlap, round to avoid numerical fluctuation
    overlap = pd.merge(rec1.round(3), rec2.round(3), how="inner")
    assert len(overlap) == 0, (
        f"Recommendations are overlapping!\n\nRecommendations 1:\n{rec1}\n\n"
        f"Recommendations 2:\n{rec2}\n\nOverlap:\n{overlap}"
    )


acqfs_non_pending = [
    a()
    for a in get_subclasses(AcquisitionFunction)
    if not a.supports_pending_experiments
]


@pytest.mark.parametrize(
    "acqf", acqfs_non_pending, ids=[a.abbreviation for a in acqfs_non_pending]
)
@pytest.mark.parametrize(
    "parameter_names",
    [
        param(_discrete_params, id="discrete"),
        param(_continuous_params, id="continuous"),
        param(_hybrid_params, id="hybrid"),
    ],
)
@pytest.mark.parametrize("n_grid_points", [5], ids=["g5"])
@pytest.mark.parametrize("batch_size", [1], ids=["b1"])
def test_invalid_acqf(searchspace, objective, batch_size, acqf, fake_measurements):
    """Test exception raised for acqfs that don't support pending experiments."""
    recommender = TwoPhaseMetaRecommender(
        recommender=BotorchRecommender(acquisition_function=acqf)
    )

    # Create fake measurements and pending experiments
    fake_pending_experiments = fake_measurements.copy()
    add_parameter_noise(fake_pending_experiments, searchspace.parameters)

    with pytest.raises(IncompatibleAcquisitionFunctionError):
        recommender.recommend(
            batch_size,
            searchspace,
            objective,
            measurements=fake_measurements,
            pending_experiments=fake_pending_experiments,
        )


@pytest.mark.parametrize(
    "parameter_names, invalid_pending_value",
    [
        param(["Categorical_1", "Num_disc_1"], "asd", id="cat_param_invalid_value"),
        param(["Categorical_1", "Num_disc_1"], 1337, id="cat_param_num"),
        param(["Categorical_1", "Num_disc_1"], np.nan, id="cat_param_nan"),
        param(["Num_disc_1", "Num_disc_2"], "asd", id="num_param_str"),
        param(["Num_disc_1", "Num_disc_2"], np.nan, id="num_param_nan"),
        param(["Custom_1", "Num_disc_2"], "asd", id="custom_param_str"),
        param(["Custom_1", "Num_disc_2"], 1337, id="custom_param_num"),
        param(["Custom_1", "Num_disc_2"], np.nan, id="custom_param_nan"),
        param(["Task", "Num_disc_1"], "asd", id="task_param_invalid_value"),
        param(["Task", "Num_disc_1"], 1337, id="task_param_num"),
        param(["Task", "Num_disc_1"], np.nan, id="task_param_nan"),
    ],
)
@pytest.mark.parametrize("n_grid_points", [5], ids=["g5"])
@pytest.mark.parametrize("batch_size", [3], ids=["b3"])
def test_invalid_input(
    searchspace,
    recommender,
    objective,
    batch_size,
    invalid_pending_value,
    parameter_names,
    fake_measurements,
):
    """Test exception raised for invalid pending experiments input."""
    # Create fake measurements and pending experiments
    fake_pending_experiments = fake_measurements.copy()
    fake_pending_experiments[parameter_names[0]] = invalid_pending_value

    with pytest.raises((ValueError, TypeError), match="parameter"):
        recommender.recommend(
            batch_size,
            searchspace,
            objective,
            measurements=fake_measurements,
            pending_experiments=fake_pending_experiments,
        )



"""Prior distributions.

The prior classes mimic classes from GPyTorch. For details on specification and
arguments see https://docs.gpytorch.ai/en/stable/priors.html.
"""

from baybe.priors.basic import (
    BetaPrior,
    GammaPrior,
    HalfCauchyPrior,
    HalfNormalPrior,
    LogNormalPrior,
    NormalPrior,
    SmoothedBoxPrior,
)

__all__ = [
    "BetaPrior",
    "GammaPrior",
    "HalfCauchyPrior",
    "HalfNormalPrior",
    "LogNormalPrior",
    "NormalPrior",
    "SmoothedBoxPrior",
]



## Example for full simulation loop using a table-based lookup mechanism

# This example shows a simulation for a direct arylation where all combinations have been measured.
# This allows us to access information about previously conducted experiments from .xlsx-files.

# This example assumes some basic familiarity with using BayBE.
# We thus refer to [`campaign`](./../Basics/campaign.md) for a basic example.

### Necessary imports for this example

import os

import pandas as pd
import seaborn as sns

from baybe import Campaign
from baybe.objectives import SingleTargetObjective
from baybe.parameters import (
    CategoricalParameter,
    NumericalDiscreteParameter,
    SubstanceParameter,
)
from baybe.recommenders import RandomRecommender
from baybe.searchspace import SearchSpace
from baybe.simulation import simulate_scenarios
from baybe.targets import NumericalTarget
from examples.utils import create_example_plots

### Parameters for a full simulation loop

# For the full simulation, we need to define some additional parameters.
# These are the number of Monte Carlo runs and the number of experiments to be conducted per run.

SMOKE_TEST = "SMOKE_TEST" in os.environ

N_DOE_ITERATIONS = 2 if SMOKE_TEST else 20
N_MC_ITERATIONS = 2 if SMOKE_TEST else 200
BATCH_SIZE = 1 if SMOKE_TEST else 2

### Lookup functionality and data creation

# We read the information about the conducted experiments from a .xlsx-file.
# This data set was obtained from [Shields, B.J., Stevens et al. Nature 590, 89–96 (2021)](https://doi.org/10.1038/s41586-021-03213-y) and contains measurements of a reaction yield,
# varying typical reaction conditions.
# Depending on your system and settings, you might need to slightly adjust the following paths.
# The reason is that it depends on the folder in which you execute the `python` call.
# This code assumes that you call `python` either from the repository root folder or this folder.

try:
    lookup = pd.read_csv("benchmarks/data/direct_arylation/data.csv")
except FileNotFoundError as e:
    lookup = pd.read_csv("../../../benchmarks/data/direct_arylation/data.csv")
    print(e)

# As usual, we set up some experiment.
# Note that we now need to ensure that the names fit the names in the provided .xlsx file!

dict_solvent = {
    "DMAc": r"CC(N(C)C)=O",
    "Butyornitrile": r"CCCC#N",
    "Butyl Ester": r"CCCCOC(C)=O",
    "p-Xylene": r"CC1=CC=C(C)C=C1",
}
dict_base = {
    "Potassium acetate": r"O=C([O-])C.[K+]",
    "Potassium pivalate": r"O=C([O-])C(C)(C)C.[K+]",
    "Cesium acetate": r"O=C([O-])C.[Cs+]",
    "Cesium pivalate": r"O=C([O-])C(C)(C)C.[Cs+]",
}
dict_ligand = {
    "BrettPhos": r"CC(C)C1=CC(C(C)C)=C(C(C(C)C)=C1)C2=C(P(C3CCCCC3)C4CCCCC4)C(OC)="
    "CC=C2OC",
    "Di-tert-butylphenylphosphine": r"CC(C)(C)P(C1=CC=CC=C1)C(C)(C)C",
    "(t-Bu)PhCPhos": r"CN(C)C1=CC=CC(N(C)C)=C1C2=CC=CC=C2P(C(C)(C)C)C3=CC=CC=C3",
    "Tricyclohexylphosphine": r"P(C1CCCCC1)(C2CCCCC2)C3CCCCC3",
    "PPh3": r"P(C1=CC=CC=C1)(C2=CC=CC=C2)C3=CC=CC=C3",
    "XPhos": r"CC(C1=C(C2=CC=CC=C2P(C3CCCCC3)C4CCCCC4)C(C(C)C)=CC(C(C)C)=C1)C",
    "P(2-furyl)3": r"P(C1=CC=CO1)(C2=CC=CO2)C3=CC=CO3",
    "Methyldiphenylphosphine": r"CP(C1=CC=CC=C1)C2=CC=CC=C2",
    "1268824-69-6": r"CC(OC1=C(P(C2CCCCC2)C3CCCCC3)C(OC(C)C)=CC=C1)C",
    "JackiePhos": r"FC(F)(F)C1=CC(P(C2=C(C3=C(C(C)C)C=C(C(C)C)C=C3C(C)C)C(OC)=CC=C2OC)"
    r"C4=CC(C(F)(F)F)=CC(C(F)(F)F)=C4)=CC(C(F)(F)F)=C1",
    "SCHEMBL15068049": r"C[C@]1(O2)O[C@](C[C@]2(C)P3C4=CC=CC=C4)(C)O[C@]3(C)C1",
    "Me2PPh": r"CP(C)C1=CC=CC=C1",
}

### Creating the Objective

objective = SingleTargetObjective(target=NumericalTarget(name="yield", mode="MAX"))

### Constructing campaigns for the simulation loop

# In this example, we create several campaigns.
# First let us create three campaigns that each use a different chemical encoding to
# treat substances.

substance_encodings = ["MORDRED", "RDKIT2DDESCRIPTORS", "ECFP"]
scenarios = {
    encoding: Campaign(
        searchspace=SearchSpace.from_product(
            parameters=[
                SubstanceParameter(
                    name="Solvent", data=dict_solvent, encoding=encoding
                ),
                SubstanceParameter(name="Base", data=dict_base, encoding=encoding),
                SubstanceParameter(name="Ligand", data=dict_ligand, encoding=encoding),
                NumericalDiscreteParameter(
                    name="Temp_C", values=[90, 105, 120], tolerance=2
                ),
                NumericalDiscreteParameter(
                    name="Concentration", values=[0.057, 0.1, 0.153]
                ),
            ]
        ),
        objective=objective,
    )
    for encoding in substance_encodings
}

# Now we create another campaign that treats the substances as simple one-hot encoded categories.

parameters = [
    CategoricalParameter(name="Solvent", values=dict_solvent.keys(), encoding="OHE"),
    CategoricalParameter(name="Base", values=dict_base.keys(), encoding="OHE"),
    CategoricalParameter(name="Ligand", values=dict_ligand.keys(), encoding="OHE"),
    NumericalDiscreteParameter(name="Temp_C", values=[90, 105, 120], tolerance=2),
    NumericalDiscreteParameter(
        name="Concentration", values=[0.057, 0.1, 0.153], tolerance=0.005
    ),
]
campaign_ohe = Campaign(
    searchspace=SearchSpace.from_product(parameters=parameters),
    objective=objective,
)

# Finally, as baseline, we specify a campaign which provides recommendations randomly.

campaign_rand = Campaign(
    searchspace=SearchSpace.from_product(parameters=parameters),
    recommender=RandomRecommender(),
    objective=objective,
)

# Update the scenarios:

scenarios.update({"OneHot": campaign_ohe, "Random Baseline": campaign_rand})

# We can now use the `simulate_scenarios` function to simulate a full optimization loop.
# Note that this function enables to run multiple scenarios by a single function call.
# For this, it is necessary to define the `scenarios` dictionary, mapping names to
# campaigns.

results = simulate_scenarios(
    scenarios,
    lookup,
    batch_size=BATCH_SIZE,
    n_doe_iterations=N_DOE_ITERATIONS,
    n_mc_iterations=N_MC_ITERATIONS,
)

# Let's visualize the results. As you can see, the type of encoding has a tremendous
# impact on the outcome, with chemical encodings performing much better than
# traditional ones at almost no extra cost.

results.rename(columns={"Scenario": "Substance Encoding"}, inplace=True)
ax = sns.lineplot(
    data=results,
    marker="o",
    markersize=10,
    x="Num_Experiments",
    y="yield_CumBest",
    hue="Substance Encoding",
)
create_example_plots(ax=ax, base_name="full_lookup")



"""Base class for all nonpredictive recommenders."""

import gc
import warnings
from abc import ABC

import pandas as pd
from attrs import define
from typing_extensions import override

from baybe.exceptions import IncompatibleArgumentError, UnusedObjectWarning
from baybe.objectives.base import Objective
from baybe.recommenders.pure.base import PureRecommender
from baybe.searchspace.core import SearchSpace


@define
class NonPredictiveRecommender(PureRecommender, ABC):
    """Abstract base class for all nonpredictive recommenders."""

    @override
    def recommend(
        self,
        batch_size: int,
        searchspace: SearchSpace,
        objective: Objective | None = None,
        measurements: pd.DataFrame | None = None,
        pending_experiments: pd.DataFrame | None = None,
    ) -> pd.DataFrame:
        if pending_experiments is not None:
            raise IncompatibleArgumentError(
                f"Pending experiments were passed to '{self.__class__.__name__}"
                f".{self.recommend.__name__}' but non-predictive recommenders "
                f"cannot use this information. If you want to exclude the pending "
                f"experiments from the candidate set, adjust the search space "
                f"accordingly."
            )
        if (measurements is not None) and not measurements.empty:
            warnings.warn(
                f"'{self.recommend.__name__}' was called with a non-empty "
                f"set of measurements but '{self.__class__.__name__}' does not "
                f"utilize any training data, meaning that the argument is ignored.",
                UnusedObjectWarning,
            )
        if objective is not None:
            warnings.warn(
                f"'{self.recommend.__name__}' was called with a an explicit objective "
                f"but '{self.__class__.__name__}' does not "
                f"consider any objectives, meaning that the argument is ignored.",
                UnusedObjectWarning,
            )
        return super().recommend(
            batch_size=batch_size,
            searchspace=searchspace,
            objective=objective,
            measurements=measurements,
            pending_experiments=None,
        )


# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Optional ngboost import."""

from baybe.exceptions import OptionalImportError

try:
    from ngboost import NGBRegressor
except ModuleNotFoundError as ex:
    raise OptionalImportError(name="ngboost") from ex

__all__ = [
    "NGBRegressor",
]



"""Conversion utilities."""

from collections.abc import Sequence
from fractions import Fraction
from typing import Any, TypeVar

from attrs import Attribute

_T = TypeVar("_T")


def fraction_to_float(value: str | float | Fraction, /) -> float:
    """Convert the provided input representing a fraction into a float.

    Args:
        value: The input to be converted.

    Returns:
        The float representation of the given input.

    Raises:
        ValueError: If the input was provided as string but could not be interpreted as
            fraction.
    """
    if isinstance(value, str):
        try:
            value = Fraction(value)
        except ValueError as err:
            raise ValueError(
                f"The provided input '{value}' could not be interpreted as a fraction."
            ) from err
    return float(value)


def nonstring_to_tuple(x: Sequence[_T], self: type, field: Attribute) -> tuple[_T, ...]:
    """Convert a sequence to tuple but raise an exception for string input."""
    if isinstance(x, str):
        raise ValueError(
            f"Argument passed to '{field.alias}' of class '{self.__class__.__name__}' "
            f"must be a sequence but cannot be a string."
        )
    return tuple(x)


def _indent(text: str, amount: int = 3, ch: str = " ") -> str:
    """Indent a given text by a certain amount."""
    padding = amount * ch
    return "".join(padding + line for line in text.splitlines(keepends=True))


def to_string(header: str, *fields: Any, single_line: bool = False) -> str:
    """Create a nested string representation.

    Args:
        header: The header, typically the name of a class.
        *fields: Fields to be printed with an indentation.
        single_line: If ``True``, print the representation on a single line.
            Only applicable when given a single field.

    Raises:
        ValueError: If ``single_line`` is ``True`` but ``fields`` contains more than one
            element.

    Returns:
        The string representation with indented fields.
    """
    if single_line:
        if len(fields) > 1:
            raise ValueError(
                "``single_line`` is only applicable when given a single field."
            )
        # Since single line headers look ugly without a ":", we add it manually
        header = header if header.endswith(":") else header + ":"
        return f"{header} {str(fields[0])}"

    return "\n".join([header] + [_indent(str(f)) for f in fields])



"""Utility for building the documentation."""

import argparse
import os
import pathlib
from subprocess import check_call, run

from build_examples import build_examples
from check_links import check_links
from utils import adjust_pictures

from baybe.telemetry import VARNAME_TELEMETRY_ENABLED

parser = argparse.ArgumentParser()
parser.add_argument(
    "-e",
    "--run-examples",
    help="Re-run the examples.",
    action="store_true",
)
parser.add_argument(
    "-l",
    "--no_linkcheck",
    help="Do not check the links.",
    action="store_true",
)
parser.add_argument(
    "-r",
    "--full-rebuild",
    help="Perform a full rebuild, independent of `-e` and `-l` flags.",
    action="store_true",
)
parser.add_argument(
    "-w",
    "--include-warnings",
    help="Include warnings when processing the examples. The default is ignoring them.",
    action="store_true",
)
parser.add_argument(
    "-f",
    "--force",
    help="Force-build the documentation, even when there are warnings or errors.",
    action="store_true",
)

# Parse input arguments
args = parser.parse_args()
RUN_EXAMPLES = args.run_examples
LINKCHECK = not args.no_linkcheck
FULL_REBUILD = args.full_rebuild
INCLUDE_WARNINGS = args.include_warnings
FORCE = args.force


def build_documentation(
    run_examples: bool = False,
    verify_links: bool = False,
    full_rebuild: bool = False,
    force: bool = False,
) -> None:
    """Build the documentation.

    A full build of the documentation consists of converting the examples into jupyter
    notebooks, executing them, transforming them into markdown files, as well as
    checking all links and performing the actual ``sphinx-build``. Such a full build can
    be triggered using the ``full_rebuild`` flag.
    If this flag is not set, this function tries to re-use as much of potentially
    existing structures like already built examples as possible. This behavior can be
    changed by using the other flags.

    Args:
        run_examples: Fully recalculate the examples. If this is ``False`` and no
            folder containing an already built set of examples is found, dummy files
            replicating the structure of the examples are created.
        verify_links: Check both internal and external links.
        full_rebuild: Perform a full rebuild of the documentation, including a
            recalculation of the examples and checking the links. Note that this option
            ignores the choices for ``run_examples`` and ``check_links`` if set to
            ``True.
        force: Force-build the steps, ignoring any errors or warnings.
    """
    examples_directory = pathlib.Path("docs/examples")
    examples_exist = examples_directory.is_dir()

    rerun_examples = run_examples or full_rebuild
    perform_linkcheck = verify_links or full_rebuild

    if rerun_examples:
        build_examples(
            destination_directory=examples_directory,
            dummy=False,
            remove_dir=examples_exist,
        )
    elif not examples_exist:
        # Perform dummy-build of examples in the case that they should not be
        # re-calculated and the folder does not exist.
        build_examples(
            destination_directory=examples_directory,
            dummy=True,
            remove_dir=examples_exist,
        )

    if perform_linkcheck:
        check_links()

    # Directory where the documentation is build.
    build_dir = pathlib.Path("docs/build")

    # The actual call that will be made to build the documentation
    building_call = [
        "sphinx-build",
        "-b",
        "html",
        "docs",
        build_dir,
        "-n",  # Being nitpicky
        "-W",  # Fail when encountering an error or a warning
    ]

    if force:
        print("Force-building the documentation, ignoring errors and warnings.")
        # In force mode, we do not want to fail, even if an error code is returned.
        # Hence, we use run instead of check_call
        run(building_call + ["--keep-going"], check=False)
    else:
        check_call(building_call)


if __name__ == "__main__":
    if not INCLUDE_WARNINGS:
        os.environ["PYTHONWARNINGS"] = "ignore"

    os.environ[VARNAME_TELEMETRY_ENABLED] = "false"

    print(f"{LINKCHECK=}")

    build_documentation(
        run_examples=RUN_EXAMPLES,
        verify_links=LINKCHECK,
        full_rebuild=FULL_REBUILD,
        force=FORCE,
    )

    # CLEANUP
    # Adjust the banner in the index and the README
    adjust_pictures(
        "docs/build/index.html",
        match="banner",
        light_version="banner2",
        dark_version="banner1",
    )
    # Adjust the chemical encoding example picture in the index and the README
    adjust_pictures(
        "docs/build/index.html",
        match="_images/full_lookup_light",
        light_version="_images/full_lookup_light",
        dark_version="_images/full_lookup_dark",
    )



"""Serialization utilities for benchmarking objects."""

import json
from datetime import datetime, timedelta
from typing import Any

import cattrs
import pandas as pd

from baybe.serialization.core import (
    _structure_dataframe_hook,
    _unstructure_dataframe_hook,
)

converter = cattrs.GenConverter(unstruct_collection_overrides={set: list})
"""The converter for benchmarking objects."""

converter.register_unstructure_hook(pd.DataFrame, _unstructure_dataframe_hook)
converter.register_structure_hook(pd.DataFrame, _structure_dataframe_hook)
converter.register_unstructure_hook(datetime, lambda x: x.isoformat())
converter.register_structure_hook(datetime, lambda x, _: datetime.fromisoformat(x))
converter.register_unstructure_hook(timedelta, lambda x: f"{x.total_seconds()}s")
converter.register_structure_hook(
    timedelta, lambda x, _: timedelta(seconds=float(x.removesuffix("s")))
)


class BenchmarkSerialization:
    """Mixin class providing serialization methods."""

    def to_dict(self) -> dict[str, Any]:
        """Create an object's dictionary representation."""
        return converter.unstructure(self)

    def to_json(self) -> str:
        """Create an object's JSON representation."""
        return json.dumps(self.to_dict())



"""Meta recommenders that switch recommenders based on the experimentation progress."""
# TODO After bayesian recommenders are enabled with no training data, a refactoring of
#  this file will resolve type errors
# mypy: disable-error-code="arg-type"

import gc
from abc import abstractmethod
from collections.abc import Iterable, Iterator
from typing import Literal

import pandas as pd
from attrs import Factory, define, field, fields
from attrs.validators import deep_iterable, ge, in_, instance_of
from typing_extensions import override

from baybe.exceptions import NoRecommendersLeftError
from baybe.objectives.base import Objective
from baybe.recommenders.base import RecommenderProtocol
from baybe.recommenders.meta.base import MetaRecommender
from baybe.recommenders.pure.bayesian.botorch import BotorchRecommender
from baybe.recommenders.pure.nonpredictive.sampling import RandomRecommender
from baybe.searchspace import SearchSpace
from baybe.serialization import (
    block_deserialization_hook,
    block_serialization_hook,
    converter,
)
from baybe.utils.conversion import to_string


@define
class TwoPhaseMetaRecommender(MetaRecommender):
    """A two-phased meta recommender that switches at a certain specified point.

    The recommender is switched when a new (batch) recommendation is requested and
    the training data set size (i.e., the total number of collected measurements
    including those gathered before the meta recommender was active) is equal to or
    greater than the number specified via the ``switch_after`` parameter.

    Note:
        Throughout each phase, the meta recommender reuses the **same** recommender
        object, that is, no new instances are created. Therefore, special attention is
        required when using the meta recommender with stateful recommenders.
    """

    initial_recommender: RecommenderProtocol = field(
        factory=RandomRecommender, validator=instance_of(RecommenderProtocol)
    )
    """The initial recommender used by the meta recommender."""

    recommender: RecommenderProtocol = field(
        factory=BotorchRecommender, validator=instance_of(RecommenderProtocol)
    )
    """The recommender used by the meta recommender after the switch."""

    switch_after: int = field(default=1, validator=[instance_of(int), ge(1)])
    """The number of experiments required for the recommender to switch."""

    remain_switched: bool = field(default=False, validator=instance_of(bool))
    """Determines if the recommender should remain switched even if the number of
    experiments falls below the threshold value in subsequent calls."""

    # TODO: These should **not** be exposed via the constructor but the workaround
    #   is currently needed for correct (de-)serialization. A proper approach would be
    #   to not set them via the constructor but through a custom hook in combination
    #   with `_cattrs_include_init_false=True`. However, the way
    #   `get_base_structure_hook` is currently designed prevents such a hook from
    #   taking action.
    _has_switched: bool = field(default=False, alias="_has_switched")
    """Indicates if the switch has already occurred."""

    @override
    @property
    def is_stateful(self) -> bool:
        return self.remain_switched

    @override
    def select_recommender(
        self,
        batch_size: int | None = None,
        searchspace: SearchSpace | None = None,
        objective: Objective | None = None,
        measurements: pd.DataFrame | None = None,
        pending_experiments: pd.DataFrame | None = None,
    ) -> RecommenderProtocol:
        n_data = len(measurements) if measurements is not None else 0
        if (n_data >= self.switch_after) or (
            self._has_switched and self.remain_switched
        ):
            self._has_switched = True
            return self.recommender

        return self.initial_recommender

    @override
    def __str__(self) -> str:
        fields = [
            to_string("Initial recommender", self.initial_recommender),
            to_string("Recommender", self.recommender),
            to_string("Switch after", self.switch_after, single_line=True),
            to_string("Remain switched", self.remain_switched, single_line=True),
            to_string("Has switched", self._has_switched, single_line=True),
        ]
        return to_string(self.__class__.__name__, *fields)


@define
class BaseSequentialMetaRecommender(MetaRecommender):
    """Base class for sequential meta recommenders."""

    # TODO: These should **not** be exposed via the constructor but the workaround
    #   is currently needed for correct (de-)serialization. A proper approach would be
    #   to not set them via the constructor but through a custom hook in combination
    #   with `_cattrs_include_init_false=True`. However, the way
    #   `get_base_structure_hook` is currently designed prevents such a hook from
    #   taking action.
    _step: int = field(default=0, alias="_step", kw_only=True)
    """An index pointing to the current position in the recommender sequence."""

    _was_used: bool = field(default=False, alias="_was_used", kw_only=True)
    """Boolean flag indicating if the current recommender has been used."""

    _n_last_measurements: int = field(
        default=0, alias="_n_last_measurements", kw_only=True
    )
    """The number of measurements available at the last successful recommend call."""

    @override
    @property
    def is_stateful(self) -> bool:
        return True

    @abstractmethod
    def _get_recommender_at_current_step(self) -> RecommenderProtocol:
        """Get the recommender at the current sequence position."""

    @override
    def select_recommender(
        self,
        batch_size: int | None = None,
        searchspace: SearchSpace | None = None,
        objective: Objective | None = None,
        measurements: pd.DataFrame | None = None,
        pending_experiments: pd.DataFrame | None = None,
    ) -> RecommenderProtocol:
        # If the training dataset size has decreased, something went wrong
        if (
            n_data := len(measurements) if measurements is not None else 0
        ) < self._n_last_measurements:
            raise RuntimeError(
                f"The training dataset size decreased from {self._n_last_measurements} "
                f"to {n_data} since the last function call, which indicates that "
                f"'{self.__class__.__name__}' was not used as intended."
            )

        # Check conditions if the next recommender in sequence is required
        more_data = n_data > self._n_last_measurements
        if (not self._was_used) or (not more_data):
            return self._get_recommender_at_current_step()

        # Move on to the next recommender
        self._step += 1
        self._was_used = False
        return self._get_recommender_at_current_step()

    @override
    def recommend(
        self,
        batch_size: int,
        searchspace: SearchSpace,
        objective: Objective | None = None,
        measurements: pd.DataFrame | None = None,
        pending_experiments: pd.DataFrame | None = None,
    ) -> pd.DataFrame:
        recommendation = super().recommend(
            batch_size, searchspace, objective, measurements, pending_experiments
        )
        self._was_used = True

        # Remember the training dataset size for the next call
        self._n_last_measurements = len(measurements) if measurements is not None else 0

        return recommendation


@define
class SequentialMetaRecommender(BaseSequentialMetaRecommender):
    """A meta recommender that uses a pre-defined sequence of recommenders.

    A new recommender is taken from the sequence whenever at least one new measurement
    is available, until all recommenders are exhausted. More precisely, a recommender
    change is triggered whenever the size of the training dataset increases; the
    actual content of the dataset is ignored.

    Note:
        The provided sequence of recommenders will be internally pre-collected into a
        list. If this is not acceptable, consider using
        :class:`baybe.recommenders.meta.sequential.StreamingSequentialMetaRecommender`
        instead.

    Raises:
        RuntimeError: If the training dataset size decreased compared to the previous
            call.
        NoRecommendersLeftError: If more recommenders are requested than there are
            recommenders available and ``mode="raise"``.
    """

    recommenders: list[RecommenderProtocol] = field(
        converter=list, validator=deep_iterable(instance_of(RecommenderProtocol))
    )
    """A finite-length sequence of recommenders to be used. For infinite-length
    iterables, see
    :class:`baybe.recommenders.meta.sequential.StreamingSequentialMetaRecommender`."""

    mode: Literal["raise", "reuse_last", "cyclic"] = field(
        default="raise",
        validator=in_(("raise", "reuse_last", "cyclic")),
    )
    """Defines what shall happen when the last recommender in the sequence has been
    consumed but additional recommender changes are triggered:

    * ``"raise"``: An error is raised.
    * ``"reuse_last"``: The last recommender in the sequence is used indefinitely.
    * ``"cycle"``: The selection restarts from the beginning of the sequence.
    """

    @override
    def _get_recommender_at_current_step(self) -> RecommenderProtocol:
        idx = self._step
        if self.mode == "reuse_last":
            idx = min(idx, len(self.recommenders) - 1)
        elif self.mode == "cyclic":
            idx %= len(self.recommenders)
        try:
            return self.recommenders[idx]
        except IndexError as ex:
            raise NoRecommendersLeftError(
                f"A total of {self._step + 1} recommender(s) was/were requested but "
                f"the provided sequence contains only {self._step} element(s). "
                f"Add more recommenders or adjust the "
                f"'{fields(SequentialMetaRecommender).mode.name}' attribute."
            ) from ex

    @override
    def __str__(self) -> str:
        fields = [
            to_string("Recommenders", self.recommenders),
            to_string("Mode", self.mode, single_line=True),
        ]
        return to_string(self.__class__.__name__, *fields)


@define
class StreamingSequentialMetaRecommender(BaseSequentialMetaRecommender):
    """A meta recommender that switches between recommenders from an iterable.

    Similar to :class:`baybe.recommenders.meta.sequential.SequentialMetaRecommender`
    but without explicit list conversion. This enables a number of advanced use cases:

    * It supports arbitrary iterables, allowing to configure recommender sequences
      of infinite length. This is useful when the total number of iterations unknown
      in advance.
    * It can be used to adaptively adjust the recommender sequence based on the
      latest context available outside the class, by modifying the iterable on the fly.

    The downside is that serialization is not supported.

    Raises:
        NoRecommendersLeftError: If more recommenders are requested than there are
            recommenders available.
    """

    recommenders: Iterable[RecommenderProtocol] = field()
    """An iterable providing the recommenders to be used."""

    _iterator: Iterator = field(
        init=False,
        default=Factory(lambda self: iter(self.recommenders), takes_self=True),
    )
    """The iterator used to traverse the recommenders."""

    _last_recommender: RecommenderProtocol | None = field(init=False, default=None)
    """The recommender returned from the last call."""

    _position_of_latest_recommender: int | None = field(init=False, default=None)
    """The position of the latest recommender fetched from the iterable."""

    @override
    def _get_recommender_at_current_step(self) -> RecommenderProtocol:
        if self._step != self._position_of_latest_recommender:
            try:
                self._last_recommender = next(self._iterator)
            except StopIteration as ex:
                raise NoRecommendersLeftError(
                    f"A total of {self._step + 1} recommender(s) was/were requested "
                    f"but the provided iterator provided only {self._step} element(s). "
                ) from ex
            self._position_of_latest_recommender = self._step

        # By now, the first recommender has been fetched
        assert self._last_recommender is not None

        return self._last_recommender

    @override
    def __str__(self) -> str:
        fields = [
            to_string("Recommenders", self.recommenders),
        ]
        return to_string(self.__class__.__name__, *fields)


# The recommender iterable cannot be serialized
converter.register_unstructure_hook(
    StreamingSequentialMetaRecommender, block_serialization_hook
)
converter.register_structure_hook(
    StreamingSequentialMetaRecommender, block_deserialization_hook
)

# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Optional import for the insights subpackage."""

from baybe.exceptions import OptionalImportError

try:
    import shap
except ModuleNotFoundError as ex:
    raise OptionalImportError(name="shap", group="insights") from ex

__all__ = [
    "shap",
]



"""Tests for insights subpackage."""

import inspect
from contextlib import nullcontext
from unittest import mock

import numpy as np
import pandas as pd
import pytest
from pytest import mark, param

from baybe._optional.info import INSIGHTS_INSTALLED
from baybe.exceptions import IncompatibleExplainerError, NoMeasurementsError
from baybe.objectives import (
    DesirabilityObjective,
    ParetoObjective,
    SingleTargetObjective,
)
from baybe.targets import NumericalTarget
from baybe.utils.basic import is_all_instance

if not INSIGHTS_INSTALLED:
    pytest.skip("Optional insights package not installed.", allow_module_level=True)


from matplotlib import pyplot as plt

from baybe import insights
from baybe._optional.insights import shap
from baybe.campaign import Campaign
from baybe.insights.shap import (
    NON_SHAP_EXPLAINERS,
    SHAP_EXPLAINERS,
    SHAP_PLOTS,
    KernelExplainer,
    SHAPInsight,
    _get_explainer_cls,
    is_shap_explainer,
)
from tests.conftest import run_iterations


@pytest.fixture
def n_grid_points():
    return 5


@pytest.fixture
def n_iterations():
    return 2


@pytest.fixture
def batch_size():
    return 2


@pytest.fixture(
    params=[
        ["Conti_finite1", "Conti_finite2"],
        ["Num_disc_1", "Fraction_1"],
        ["Num_disc_1", "Conti_finite1"],
        ["Num_disc_1", "Categorical_1"],
        ["Conti_finite1", "Categorical_1"],
    ],
    ids=[
        "params_conti",
        "params_disc_num",
        "params_hybrid_num",
        "params_disc_cat",
        "params_hybrid_cat",
    ],
)
def parameter_names(request):
    return request.param


def _has_required_init_parameters(cls: type[shap.Explainer]) -> bool:
    """Check if non-shap initializer has required standard parameters."""
    REQUIRED_PARAMETERS = ["self", "model", "data"]
    init_signature = inspect.signature(cls.__init__)
    parameters = list(init_signature.parameters.keys())
    return parameters[:3] == REQUIRED_PARAMETERS


@pytest.mark.parametrize("explainer_name", NON_SHAP_EXPLAINERS)
def test_non_shap_signature(explainer_name):
    """Non-SHAP explainers must have the required signature."""
    assert _has_required_init_parameters(_get_explainer_cls(explainer_name))


def _test_shap_insight(campaign, explainer_cls, use_comp_rep, is_shap):
    """Helper function for general SHAP explainer tests."""
    context = nullcontext()
    if (
        (not use_comp_rep)
        and (explainer_cls != "KernelExplainer")
        and any(not p.is_numerical for p in campaign.parameters)
    ):
        # We expect a validation error in case an explanation with an unsupported
        # explainer type is attempted on a search space representation with
        # non-numerical entries
        context = pytest.raises(IncompatibleExplainerError)

    with context:
        shap_insight = SHAPInsight.from_campaign(
            campaign,
            explainer_cls=explainer_cls,
            use_comp_rep=use_comp_rep,
        )

        # Sanity check explainers
        assert isinstance(shap_insight, insights.SHAPInsight)
        assert all(is_shap_explainer(e) == is_shap for e in shap_insight.explainers)

        # Sanity check explanation
        df = campaign.measurements[[p.name for p in campaign.parameters]]
        if use_comp_rep:
            df = campaign.searchspace.transform(df)
        shap_explanations = shap_insight.explain(df)
        assert isinstance(shap_explanations, tuple)
        assert is_all_instance(shap_explanations, shap.Explanation)


@mark.slow
@mark.parametrize(
    "objective",
    [
        param(SingleTargetObjective(NumericalTarget("t1", "MAX")), id="single"),
        param(
            DesirabilityObjective(
                (
                    NumericalTarget("t1", mode="MAX", bounds=(0, 100)),
                    NumericalTarget("t2", mode="MIN", bounds=(0, 100)),
                )
            ),
            id="desirability",
        ),
        param(
            ParetoObjective(
                (NumericalTarget("t1", mode="MAX"), NumericalTarget("t2", mode="MIN"))
            ),
            id="pareto",
        ),
    ],
)
@mark.parametrize("explainer_cls", SHAP_EXPLAINERS)
@mark.parametrize("use_comp_rep", [False, True], ids=["exp", "comp"])
def test_shap_explainers(ongoing_campaign, explainer_cls, use_comp_rep):
    """Test the explain functionalities with measurements."""
    _test_shap_insight(ongoing_campaign, explainer_cls, use_comp_rep, is_shap=True)


@mark.parametrize("explainer_cls", NON_SHAP_EXPLAINERS)
def test_non_shap_explainers(ongoing_campaign, explainer_cls):
    """Test the non-SHAP explainers in computational representation."""
    _test_shap_insight(
        ongoing_campaign, explainer_cls, use_comp_rep=True, is_shap=False
    )


@mark.slow
@mark.parametrize("explainer_cls", ["KernelExplainer"], ids=["KernelExplainer"])
@mark.parametrize("use_comp_rep", [False, True], ids=["exp", "comp"])
def test_invalid_explained_data(ongoing_campaign, explainer_cls, use_comp_rep):
    """Test invalid explained data."""
    shap_insight = SHAPInsight.from_campaign(
        ongoing_campaign,
        explainer_cls=explainer_cls,
        use_comp_rep=use_comp_rep,
    )
    df = pd.DataFrame({"Num_disc_1": [0, 2]})
    with pytest.raises(ValueError, match="must contain all columns that were used"):
        shap_insight.explain(df)


@mark.slow
@mark.parametrize("use_comp_rep", [False, True], ids=["exp", "comp"])
@mark.parametrize("plot_type", SHAP_PLOTS)
def test_plots(ongoing_campaign: Campaign, use_comp_rep, plot_type):
    """Test the default SHAP plots."""
    shap_insight = SHAPInsight.from_campaign(
        ongoing_campaign,
        use_comp_rep=use_comp_rep,
    )
    df = ongoing_campaign.measurements[[p.name for p in ongoing_campaign.parameters]]
    if use_comp_rep:
        df = ongoing_campaign.searchspace.transform(df)

    with mock.patch("matplotlib.pyplot.show"):
        shap_insight.plot(plot_type, df)
        plt.close()


def test_updated_campaign_explanations(campaign, n_iterations, batch_size):
    """Test explanations for campaigns with updated measurements."""
    with pytest.raises(
        NoMeasurementsError,
        match="The campaign does not contain any measurements.",
    ):
        SHAPInsight.from_campaign(campaign)

    run_iterations(campaign, n_iterations=n_iterations, batch_size=batch_size)
    shap_insight = SHAPInsight.from_campaign(campaign)
    df = campaign.measurements[[p.name for p in campaign.parameters]]
    explanation_1 = shap_insight.explain(df)

    run_iterations(campaign, n_iterations=n_iterations, batch_size=batch_size)
    shap_insight = SHAPInsight.from_campaign(campaign)
    df = campaign.measurements[[p.name for p in campaign.parameters]]
    explanation_2 = shap_insight.explain(df)

    assert explanation_1 != explanation_2, "SHAP explanations should not be identical."


def test_creation_from_recommender(ongoing_campaign):
    """Test the creation of SHAP insights from a recommender."""
    recommender = ongoing_campaign.recommender.recommender
    shap_insight = SHAPInsight.from_recommender(
        recommender,
        ongoing_campaign.searchspace,
        ongoing_campaign.objective,
        ongoing_campaign.measurements,
    )
    assert isinstance(shap_insight, insights.SHAPInsight)


def test_column_permutation():
    """Explaining data with permuted columns gives permuted explanations."""
    N = 10

    # Create insights object and test data
    background_data = pd.DataFrame(np.random.random((N, 3)), columns=["x", "y", "z"])
    explainer = KernelExplainer(lambda x: x, background_data)
    insight = SHAPInsight([explainer], background_data)
    df = pd.DataFrame(np.random.random((N, 3)), columns=["x", "y", "z"])

    # Regular column order
    ex1 = insight.explain(df)

    # Permuted column order
    ex2 = insight.explain(df[["z", "x", "y"]])

    assert np.array_equal(ex1[0].values, ex2[0][:, [1, 2, 0]].values)



"""Available acquisition functions."""

import gc
import math
from abc import ABC
from typing import ClassVar

import numpy as np
import numpy.typing as npt
import pandas as pd
from attr.converters import optional as optional_c
from attr.validators import optional as optional_v
from attrs import define, field, fields
from attrs.validators import gt, instance_of, le
from typing_extensions import override

from baybe.acquisition.base import AcquisitionFunction
from baybe.searchspace import SearchSpace
from baybe.utils.basic import classproperty, convert_to_float
from baybe.utils.sampling_algorithms import DiscreteSamplingMethod, sample_numerical_df
from baybe.utils.validation import finite_float


########################################################################################
### Active Learning
@define(frozen=True)
class qNegIntegratedPosteriorVariance(AcquisitionFunction):
    """Monte Carlo based negative integrated posterior variance.

    This is typically used for active learning as it is a measure for global model
    uncertainty.
    """

    abbreviation: ClassVar[str] = "qNIPV"

    sampling_n_points: int | None = field(
        validator=optional_v([instance_of(int), gt(0)]),
        default=None,
    )
    """Number of data points sampled for integrating the posterior.

    Cannot be used if `sampling_fraction` is not `None`."""

    sampling_fraction: float | None = field(
        converter=optional_c(float),
        validator=optional_v([gt(0.0), le(1.0)]),
    )
    """Fraction of data sampled for integrating the posterior.

    Cannot be used if `sampling_n_points` is not `None`."""

    sampling_method: DiscreteSamplingMethod = field(
        converter=DiscreteSamplingMethod, default=DiscreteSamplingMethod.Random
    )
    """Sampling strategy used for integrating the posterior."""

    @sampling_fraction.default
    def _default_sampling_fraction(self) -> float | None:
        """If no sampling quantities are provided, use all points by default."""
        return 1.0 if self.sampling_n_points is None else None

    @sampling_fraction.validator
    def _validate_sampling_fraction(self, attr, value) -> None:
        """If both sampling quantities are specified, raise an error."""
        if None not in (self.sampling_fraction, self.sampling_n_points):
            raise ValueError(
                f"For '{self.__class__.__name__}', the attributes '{attr.name}' and "
                f"'{fields(self.__class__).sampling_n_points.name}' cannot "
                f"be specified at the same time."
            )

    @override
    @classproperty
    def _non_botorch_attrs(cls) -> tuple[str, ...]:
        flds = fields(qNegIntegratedPosteriorVariance)
        return (
            flds.sampling_n_points.name,
            flds.sampling_method.name,
            flds.sampling_fraction.name,
        )

    def get_integration_points(self, searchspace: SearchSpace) -> pd.DataFrame:
        """Sample points from a search space for integration purposes.

        Sampling of the discrete part can be controlled via 'sampling_method', but
        sampling of the continuous part will always be random.

        Args:
            searchspace: The searchspace from which to sample integration points.

        Returns:
            The sampled data points.

        Raises:
            ValueError: If the search space is purely continuous and
                'sampling_n_points' was not provided.
        """
        # TODO: Move the core logic to `SearchSpace` and ``Subspace*`` classes

        sampled_parts: list[pd.DataFrame] = []
        n_candidates: int | None = None

        # Discrete part
        if not searchspace.discrete.is_empty:
            candidates_discrete = searchspace.discrete.comp_rep
            n_candidates = self.sampling_n_points or math.ceil(
                self.sampling_fraction * len(candidates_discrete)  # type: ignore[operator]
            )

            sampled_disc = sample_numerical_df(
                candidates_discrete, n_candidates, method=self.sampling_method
            )

            sampled_parts.append(sampled_disc)

        # Continuous part
        if not searchspace.continuous.is_empty:
            # If a discrete part has resulted in a particular choice for n_candidates,
            # take it. Otherwise, use the user specified number of points.
            if (n_candidates := n_candidates or self.sampling_n_points) is None:
                raise ValueError(
                    f"'{fields(self.__class__).sampling_n_points.name}' must be "
                    f"provided for '{self.__class__.__name__}' when sampling purely "
                    f"continuous search spaces."
                )
            sampled_conti = searchspace.continuous.sample_uniform(n_candidates)

            # Align indices if discrete part is present
            if len(sampled_parts) > 0:
                sampled_conti.index = sampled_parts[0].index
            sampled_parts.append(sampled_conti)

        # Combine different search space parts
        result = pd.concat(sampled_parts, axis=1)

        return result


########################################################################################
### Knowledge Gradient
@define(frozen=True)
class qKnowledgeGradient(AcquisitionFunction):
    """Monte Carlo based knowledge gradient.

    This acquisition function currently only supports purely continuous spaces.
    """

    abbreviation: ClassVar[str] = "qKG"

    num_fantasies: int = field(validator=[instance_of(int), gt(0)], default=128)
    """Number of fantasies to draw for approximating the knowledge gradient.

    More samples result in a better approximation, at the expense of both increased
    memory footprint and wall time."""


########################################################################################
### Posterior Statistics
@define(frozen=True)
class PosteriorMean(AcquisitionFunction):
    """Posterior mean."""

    abbreviation: ClassVar[str] = "PM"


@define(frozen=True)
class PosteriorStandardDeviation(AcquisitionFunction):
    """Posterior standard deviation."""

    abbreviation: ClassVar[str] = "PSTD"

    maximize: bool = field(default=True, validator=instance_of(bool))
    """If ``True``, points with maximum posterior standard deviation are selected.
    If ``False``, the acquisition function value is negated, yielding a selection
    with minimal posterior standard deviation."""


@define(frozen=True)
class qPosteriorStandardDeviation(AcquisitionFunction):
    """Monte Carlo based posterior standard deviation."""

    abbreviation: ClassVar[str] = "qPSTD"


########################################################################################
### Simple Regret
@define(frozen=True)
class qSimpleRegret(AcquisitionFunction):
    """Monte Carlo based simple regret."""

    abbreviation: ClassVar[str] = "qSR"


########################################################################################
### Expected Improvement
@define(frozen=True)
class ExpectedImprovement(AcquisitionFunction):
    """Analytical expected improvement."""

    abbreviation: ClassVar[str] = "EI"


@define(frozen=True)
class qExpectedImprovement(AcquisitionFunction):
    """Monte Carlo based expected improvement."""

    abbreviation: ClassVar[str] = "qEI"


@define(frozen=True)
class LogExpectedImprovement(AcquisitionFunction):
    """Logarithmic analytical expected improvement."""

    abbreviation: ClassVar[str] = "LogEI"


@define(frozen=True)
class qLogExpectedImprovement(AcquisitionFunction):
    """Logarithmic Monte Carlo based expected improvement."""

    abbreviation: ClassVar[str] = "qLogEI"


@define(frozen=True)
class qNoisyExpectedImprovement(AcquisitionFunction):
    """Monte Carlo based noisy expected improvement."""

    abbreviation: ClassVar[str] = "qNEI"

    prune_baseline: bool = field(default=True, validator=instance_of(bool))
    """Auto-prune candidates that are unlikely to be the best."""


@define(frozen=True)
class qLogNoisyExpectedImprovement(AcquisitionFunction):
    """Logarithmic Monte Carlo based noisy expected improvement."""

    abbreviation: ClassVar[str] = "qLogNEI"

    prune_baseline: bool = field(default=True, validator=instance_of(bool))
    """Auto-prune candidates that are unlikely to be the best."""


########################################################################################
### Probability of Improvement
@define(frozen=True)
class ProbabilityOfImprovement(AcquisitionFunction):
    """Analytical probability of improvement."""

    abbreviation: ClassVar[str] = "PI"


@define(frozen=True)
class qProbabilityOfImprovement(AcquisitionFunction):
    """Monte Carlo based probability of improvement."""

    abbreviation: ClassVar[str] = "qPI"


########################################################################################
### Upper Confidence Bound
@define(frozen=True)
class UpperConfidenceBound(AcquisitionFunction):
    """Analytical upper confidence bound."""

    abbreviation: ClassVar[str] = "UCB"

    beta: float = field(converter=float, validator=finite_float, default=0.2)
    """Trade-off parameter for mean and variance.

    * ``beta > 0``: Rewards uncertainty, takes more risk.
      Limit ``inf``: Pure exploration
    * ``beta < 0``: Punishes uncertainty, takes less risk.
      Limit ``-inf``: Pure exploitation
    * ``beta = 0``: Discards knowledge about uncertainty, i.e. neither rewards nor
      punishes it, is risk-neutral.
    """


@define(frozen=True)
class qUpperConfidenceBound(AcquisitionFunction):
    """Monte Carlo based upper confidence bound."""

    abbreviation: ClassVar[str] = "qUCB"

    beta: float = field(converter=float, validator=finite_float, default=0.2)
    """See :paramref:`UpperConfidenceBound.beta`."""


########################################################################################
### ThompsonSampling
@define(frozen=True)
class qThompsonSampling(qSimpleRegret):
    """Thomson sampling, implemented via simple regret. Inherently Monte Carlo based.

    This implementation exploits the fact that one-sample-based Thompson sampling
    (i.e. where the action probability is approximated using a single posterior sample)
    is equivalent to optimizing the Monte Carlo approximated posterior mean with
    sample size one. The latter can be achieved via `qSimpleRegret` and controlling
    its sample shape attribute.
    """

    abbreviation: ClassVar[str] = "qTS"

    n_mc_samples: int = field(default=1, init=False)
    """Number of Monte Carlo samples drawn from the posterior at each design point.

    Restring the the sample size to one allows us to emulate (one-sample based)
    Thompson sampling using the regular acquisition function machinery.
    """

    @override
    @classproperty
    def _non_botorch_attrs(cls) -> tuple[str, ...]:
        flds = fields(qThompsonSampling)
        return (flds.n_mc_samples.name,)

    @override
    @classproperty
    def supports_batching(cls) -> bool:
        return False


########################################################################################
### Pareto Efficient Global Optimization (Chebyshev scalarization of targets)
@define(frozen=True)
class qLogNParEGO(AcquisitionFunction):
    """Pareto optimization via Chebyshev scalarization of the targets."""

    abbreviation: ClassVar[str] = "qLogNParEGO"
    supports_multi_output: ClassVar[bool] = True

    prune_baseline: bool = field(default=True, validator=instance_of(bool))
    """Auto-prune candidates that are unlikely to be the best."""


########################################################################################
### Hypervolume Improvement
@define(frozen=True)
class _ExpectedHypervolumeImprovement(AcquisitionFunction, ABC):
    """Expected hypervolume improvement base class."""

    supports_multi_output: ClassVar[bool] = True

    reference_point: float | tuple[float, ...] | None = field(
        default=None, converter=optional_c(convert_to_float)
    )
    """The reference point for computing the hypervolume improvement.

    * When omitted, a default reference point is computed based on the provided data.
    * When specified as a float, the value is interpreted as a multiplicative factor
      determining the reference point location based on the difference between the best
      and worst target configuration in the provided data.
    * When specified as an iterable, the contained values are directly interpreted as
      the coordinates of the reference point.
    """

    prune_baseline: bool = field(default=True, validator=instance_of(bool))
    """Auto-prune candidates that are unlikely to be the best."""

    @override
    @classproperty
    def _non_botorch_attrs(cls) -> tuple[str, ...]:
        # While BoTorch's acquisition function also expects a `ref_point` argument,
        # the attribute defined here is more general and can hence not be directly
        # matched. Thus, we bypass the auto-matching mechanism and handle it manually.
        flds = fields(qLogNoisyExpectedHypervolumeImprovement)
        return (flds.reference_point.name,)

    @staticmethod
    def compute_ref_point(
        array: npt.ArrayLike, maximize: npt.ArrayLike, factor: float = 0.1
    ) -> np.ndarray:
        """Compute a reference point for a given set of target configurations.

        The reference point is positioned relative to the worst point in the direction
        coming from the best point:

        * A factor of 0.0 results in the reference point being the worst point.
        * A factor > 0.0 moves the reference point further away from both worst and best
          points. (A factor of 1.0 exactly mirrors the best around the worst point.)
        * A factor < 0.0 moves the reference point closer to the best point.
          (A factor of -1.0 exactly places the reference point onto the best point.)

        Example:
            >>> from baybe.acquisition import qLogNEHVI

            >>> qLogNEHVI.compute_ref_point([[0, 10], [2, 20]], [True, True], 0.1)
            array([-0.2,  9. ])

            >>> qLogNEHVI.compute_ref_point([[0, 10], [2, 20]], [True, False], 0.2)
            array([-0.4, 22. ])

        Args:
            array: A 2-D array-like where each row represents a target configuration.
            maximize: A 1-D boolean array indicating which targets are to be maximized.
            factor: A numeric value controlling the location of the reference point.

        Raises:
            ValueError: If the given target configuration array is not two-dimensional.
            ValueError: If the given Boolean array is not one-dimensional.

        Returns:
            The computed reference point.
        """
        if np.ndim(array) != 2:
            raise ValueError(
                "The specified data array must have exactly two dimensions."
            )
        if np.ndim(maximize) != 1:
            raise ValueError(
                "The specified Boolean array must have exactly one dimension."
            )

        # Convert arrays
        array = np.asarray(array)
        maximize = np.where(maximize, 1.0, -1.0)

        # Compute bounds
        array = array * maximize[None, :]
        min = np.min(array, axis=0)
        max = np.max(array, axis=0)

        return (min - factor * (max - min)) * maximize


@define(frozen=True)
class qNoisyExpectedHypervolumeImprovement(_ExpectedHypervolumeImprovement):
    """Monte Carlo based noisy expected hypervolume improvement."""

    abbreviation: ClassVar[str] = "qNEHVI"


@define(frozen=True)
class qLogNoisyExpectedHypervolumeImprovement(_ExpectedHypervolumeImprovement):
    """Logarithmic Monte Carlo based noisy expected hypervolume improvement."""

    abbreviation: ClassVar[str] = "qLogNEHVI"


# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Base classes for all meta recommenders."""

import gc
from abc import ABC, abstractmethod
from typing import Any

import pandas as pd
from attrs import define
from typing_extensions import override

from baybe.exceptions import DeprecationError
from baybe.objectives.base import Objective
from baybe.recommenders.base import RecommenderProtocol
from baybe.recommenders.pure.base import PureRecommender
from baybe.recommenders.pure.nonpredictive.base import NonPredictiveRecommender
from baybe.searchspace import SearchSpace
from baybe.serialization import SerialMixin, converter, unstructure_base
from baybe.serialization.core import get_base_structure_hook


@define
class MetaRecommender(SerialMixin, RecommenderProtocol, ABC):
    """Abstract base class for all meta recommenders."""

    @property
    @abstractmethod
    def is_stateful(self) -> bool:
        """Boolean indicating if the meta recommender is stateful."""

    @abstractmethod
    def select_recommender(
        self,
        batch_size: int | None = None,
        searchspace: SearchSpace | None = None,
        objective: Objective | None = None,
        measurements: pd.DataFrame | None = None,
        pending_experiments: pd.DataFrame | None = None,
    ) -> RecommenderProtocol:
        """Select a recommender for the given experimentation context.

        See :meth:`baybe.recommenders.base.RecommenderProtocol.recommend` for details
        on the method arguments.
        """

    def get_non_meta_recommender(
        self,
        batch_size: int | None = None,
        searchspace: SearchSpace | None = None,
        objective: Objective | None = None,
        measurements: pd.DataFrame | None = None,
        pending_experiments: pd.DataFrame | None = None,
    ) -> RecommenderProtocol:
        """Follow the meta recommender chain to the selected non-meta recommender.

        Recursively calls :meth:`MetaRecommender.select_recommender` until a
        non-meta recommender is encountered, which is then returned.
        Effectively, this extracts the recommender responsible for generating
        the recommendations for the specified context.

        See :meth:`baybe.recommenders.base.RecommenderProtocol.recommend` for details
        on the method arguments.
        """
        recommender: MetaRecommender | RecommenderProtocol = self
        while isinstance(recommender, MetaRecommender):
            recommender = recommender.select_recommender(
                batch_size, searchspace, objective, measurements, pending_experiments
            )
        return recommender

    def get_current_recommender(self) -> PureRecommender:
        """Deprecated! Use :meth:`select_recommender` or
        :meth:`get_non_meta_recommender` instead.
        """  # noqa
        raise DeprecationError(
            f"'{MetaRecommender.__name__}.get_current_recommender' has been deprecated."
            f"Use '{MetaRecommender.__name__}.{self.select_recommender.__name__}' or "
            f"'{MetaRecommender.__name__}.{self.get_non_meta_recommender.__name__}' "
            f"instead."
        )

    def get_next_recommender(self) -> PureRecommender:
        """Deprecated! Use :meth:`select_recommender` or
        :meth:`get_non_meta_recommender` instead.
        """  # noqa
        raise DeprecationError(
            f"'{MetaRecommender.__name__}.get_current_recommender' has been deprecated."
            f"Use '{MetaRecommender.__name__}.{self.select_recommender.__name__}' or "
            f"'{MetaRecommender.__name__}.{self.get_non_meta_recommender.__name__}' "
            f"instead."
        )

    @override
    def recommend(
        self,
        batch_size: int,
        searchspace: SearchSpace,
        objective: Objective | None = None,
        measurements: pd.DataFrame | None = None,
        pending_experiments: pd.DataFrame | None = None,
    ) -> pd.DataFrame:
        """See :meth:`baybe.recommenders.base.RecommenderProtocol.recommend`."""
        recommender = self.select_recommender(
            batch_size=batch_size,
            searchspace=searchspace,
            objective=objective,
            measurements=measurements,
            pending_experiments=pending_experiments,
        )

        # Non-predictive recommenders should not be called with an objective or
        # measurements. Using dict value type Any here due to known mypy complication:
        # https://github.com/python/mypy/issues/5382
        optional_args: dict[str, Any] = (
            {}
            if isinstance(recommender, NonPredictiveRecommender)
            else {
                "objective": objective,
                "measurements": measurements,
            }
        )

        return recommender.recommend(
            batch_size=batch_size,
            searchspace=searchspace,
            pending_experiments=pending_experiments,
            **optional_args,
        )


# Register (un-)structure hooks
converter.register_unstructure_hook(MetaRecommender, unstructure_base)
converter.register_structure_hook(
    MetaRecommender, get_base_structure_hook(MetaRecommender)
)

# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Botorch recommender."""

from __future__ import annotations

import gc
import math
import warnings
from collections.abc import Collection, Iterable
from typing import TYPE_CHECKING, Any, ClassVar

import numpy as np
import pandas as pd
from attrs import define, field
from attrs.converters import optional as optional_c
from attrs.validators import ge, gt, instance_of
from typing_extensions import override

from baybe.acquisition.acqfs import qThompsonSampling
from baybe.constraints import ContinuousCardinalityConstraint
from baybe.constraints.utils import is_cardinality_fulfilled
from baybe.exceptions import (
    IncompatibilityError,
    IncompatibleAcquisitionFunctionError,
    InfeasibilityError,
    MinimumCardinalityViolatedWarning,
)
from baybe.parameters.numerical import _FixedNumericalContinuousParameter
from baybe.recommenders.pure.bayesian.base import BayesianRecommender
from baybe.searchspace import (
    SearchSpace,
    SearchSpaceType,
    SubspaceContinuous,
    SubspaceDiscrete,
)
from baybe.utils.conversion import to_string
from baybe.utils.dataframe import to_tensor
from baybe.utils.sampling_algorithms import (
    DiscreteSamplingMethod,
    sample_numerical_df,
)

if TYPE_CHECKING:
    from torch import Tensor


@define(kw_only=True)
class BotorchRecommender(BayesianRecommender):
    """A pure recommender utilizing Botorch's optimization machinery.

    This recommender makes use of Botorch's ``optimize_acqf_discrete``,
    ``optimize_acqf`` and ``optimize_acqf_mixed`` functions to optimize discrete,
    continuous and hybrid search spaces, respectively. Accordingly, it can be applied to
    all kinds of search spaces.

    Note:
        In hybrid search spaces, the used algorithm performs a brute-force optimization
        that can be computationally expensive. Thus, the behavior of the algorithm in
        hybrid search spaces can be controlled via two additional parameters.
    """

    # Class variables
    compatibility: ClassVar[SearchSpaceType] = SearchSpaceType.HYBRID
    # See base class.

    # Object variables
    sequential_continuous: bool = field(default=True)
    """Flag defining whether to apply sequential greedy or batch optimization in
    **continuous** search spaces. In discrete/hybrid spaces, sequential greedy
    optimization is applied automatically.
    """

    hybrid_sampler: DiscreteSamplingMethod | None = field(
        converter=optional_c(DiscreteSamplingMethod), default=None
    )
    """Strategy used for sampling the discrete subspace when performing hybrid search
    space optimization."""

    sampling_percentage: float = field(default=1.0)
    """Percentage of discrete search space that is sampled when performing hybrid search
    space optimization. Ignored when ``hybrid_sampler="None"``."""

    n_restarts: int = field(validator=[instance_of(int), gt(0)], default=10)
    """Number of times gradient-based optimization is restarted from different initial
    points. **Does not affect purely discrete optimization**.
    """

    n_raw_samples: int = field(validator=[instance_of(int), gt(0)], default=64)
    """Number of raw samples drawn for the initialization heuristic in gradient-based
    optimization. **Does not affect purely discrete optimization**.
    """

    max_n_subspaces: int = field(default=10, validator=[instance_of(int), ge(1)])
    """Threshold defining the maximum number of subspaces to consider for exhaustive
    search in the presence of cardinality constraints. If the combinatorial number of
    groupings into active and inactive parameters dictated by the constraints is greater
    than this number, that many randomly selected combinations are selected for
    optimization."""

    @sampling_percentage.validator
    def _validate_percentage(  # noqa: DOC101, DOC103
        self, _: Any, value: float
    ) -> None:
        """Validate that the given value is in fact a percentage.

        Raises:
            ValueError: If ``value`` is not between 0 and 1.
        """
        if not 0 <= value <= 1:
            raise ValueError(
                f"Hybrid sampling percentage needs to be between 0 and 1 but is {value}"
            )

    @override
    def __str__(self) -> str:
        fields = [
            to_string("Surrogate", self._surrogate_model),
            to_string(
                "Acquisition function", self.acquisition_function, single_line=True
            ),
            to_string("Compatibility", self.compatibility, single_line=True),
            to_string(
                "Sequential continuous", self.sequential_continuous, single_line=True
            ),
            to_string("Hybrid sampler", self.hybrid_sampler, single_line=True),
            to_string(
                "Sampling percentage", self.sampling_percentage, single_line=True
            ),
        ]
        return to_string(self.__class__.__name__, *fields)

    @override
    def _recommend_discrete(
        self,
        subspace_discrete: SubspaceDiscrete,
        candidates_exp: pd.DataFrame,
        batch_size: int,
    ) -> pd.Index:
        """Generate recommendations from a discrete search space.

        Args:
            subspace_discrete: The discrete subspace from which to generate
                recommendations.
            candidates_exp: The experimental representation of all discrete candidate
                points to be considered.
            batch_size: The size of the recommendation batch.

        Raises:
            IncompatibleAcquisitionFunctionError: If a non-Monte Carlo acquisition
                function is used with a batch size > 1.

        Returns:
            The dataframe indices of the recommended points in the provided
            experimental representation.
        """
        assert self._objective is not None
        acqf = self._get_acquisition_function(self._objective)
        if batch_size > 1 and not acqf.supports_batching:
            raise IncompatibleAcquisitionFunctionError(
                f"The '{self.__class__.__name__}' only works with Monte Carlo "
                f"acquisition functions for batch sizes > 1."
            )
        if batch_size > 1 and isinstance(acqf, qThompsonSampling):
            raise IncompatibilityError(
                "Thompson sampling currently only supports a batch size of 1."
            )

        from botorch.optim import optimize_acqf_discrete

        # determine the next set of points to be tested
        candidates_comp = subspace_discrete.transform(candidates_exp)
        points, _ = optimize_acqf_discrete(
            self._botorch_acqf, batch_size, to_tensor(candidates_comp)
        )

        # retrieve the index of the points from the input dataframe
        # IMPROVE: The merging procedure is conceptually similar to what
        #   `SearchSpace._match_measurement_with_searchspace_indices` does, though using
        #   a simpler matching logic. When refactoring the SearchSpace class to
        #   handle continuous parameters, a corresponding utility could be extracted.
        idxs = pd.Index(
            pd.merge(
                pd.DataFrame(points, columns=candidates_comp.columns),
                candidates_comp.reset_index(),
                on=list(candidates_comp),
                how="left",
            )["index"]
        )

        return idxs

    @override
    def _recommend_continuous(
        self,
        subspace_continuous: SubspaceContinuous,
        batch_size: int,
    ) -> pd.DataFrame:
        """Generate recommendations from a continuous search space.

        Args:
            subspace_continuous: The continuous subspace from which to generate
                recommendations.
            batch_size: The size of the recommendation batch.

        Raises:
            IncompatibleAcquisitionFunctionError: If a non-Monte Carlo acquisition
                function is used with a batch size > 1.

        Returns:
            A dataframe containing the recommendations as individual rows.
        """
        assert self._objective is not None
        if (
            batch_size > 1
            and not self._get_acquisition_function(self._objective).supports_batching
        ):
            raise IncompatibleAcquisitionFunctionError(
                f"The '{self.__class__.__name__}' only works with Monte Carlo "
                f"acquisition functions for batch sizes > 1."
            )

        points, _ = self._recommend_continuous_torch(subspace_continuous, batch_size)

        return pd.DataFrame(points, columns=subspace_continuous.parameter_names)

    def _recommend_continuous_torch(
        self, subspace_continuous: SubspaceContinuous, batch_size: int
    ) -> tuple[Tensor, Tensor]:
        """Dispatcher selecting the continuous optimization routine."""
        if subspace_continuous.constraints_cardinality:
            return self._recommend_continuous_with_cardinality_constraints(
                subspace_continuous, batch_size
            )
        else:
            return self._recommend_continuous_without_cardinality_constraints(
                subspace_continuous, batch_size
            )

    def _recommend_continuous_with_cardinality_constraints(
        self,
        subspace_continuous: SubspaceContinuous,
        batch_size: int,
    ) -> tuple[Tensor, Tensor]:
        """Recommend from a continuous search space with cardinality constraints.

        This is achieved by considering the individual restricted subspaces that can be
        obtained by splitting the parameters into sets of active and inactive
        parameters, according to what is allowed by the cardinality constraints.

        The specific collection of subspaces considered by the recommender is obtained
        as either the full combinatorial set of possible parameter splits or a random
        selection thereof, depending on the upper bound specified by the corresponding
        recommender attribute.

        In each of these spaces, the (in)activity assignment is fixed, so that the
        cardinality constraints can be removed and a regular optimization can be
        performed. The recommendation is then constructed from the combined optimization
        results of the unconstrained spaces.

        Args:
            subspace_continuous: The continuous subspace from which to generate
                recommendations.
            batch_size: The size of the recommendation batch.

        Returns:
            The recommendations and corresponding acquisition values.

        Raises:
            ValueError: If the continuous search space has no cardinality constraints.
        """
        if not subspace_continuous.constraints_cardinality:
            raise ValueError(
                f"'{self._recommend_continuous_with_cardinality_constraints.__name__}' "
                f"expects a subspace with constraints of type "
                f"'{ContinuousCardinalityConstraint.__name__}'. "
            )

        # Determine search scope based on number of inactive parameter combinations
        exhaustive_search = (
            subspace_continuous.n_inactive_parameter_combinations
            <= self.max_n_subspaces
        )
        iterator: Iterable[Collection[str]]
        if exhaustive_search:
            # If manageable, evaluate all combinations of inactive parameters
            iterator = subspace_continuous.inactive_parameter_combinations()
        else:
            # Otherwise, draw a random subset of inactive parameter combinations
            iterator = subspace_continuous._sample_inactive_parameters(
                self.max_n_subspaces
            )

        # Create iterable of subspaces to be optimized
        subspaces = (
            (subspace_continuous._enforce_cardinality_constraints(inactive_parameters))
            for inactive_parameters in iterator
        )

        points, acqf_value = self._optimize_continuous_subspaces(subspaces, batch_size)

        # Check if any minimum cardinality constraints are violated
        if not is_cardinality_fulfilled(
            pd.DataFrame(points, columns=subspace_continuous.parameter_names),
            subspace_continuous,
            check_maximum=False,
        ):
            warnings.warn(
                "At least one minimum cardinality constraint has been violated. "
                "This may occur when parameter ranges extend beyond zero in both "
                "directions, making the feasible region non-convex. For such "
                "parameters, minimum cardinality constraints are currently not "
                "enforced due to the complexity of the resulting optimization problem.",
                MinimumCardinalityViolatedWarning,
            )

        return points, acqf_value

    def _recommend_continuous_without_cardinality_constraints(
        self,
        subspace_continuous: SubspaceContinuous,
        batch_size: int,
    ) -> tuple[Tensor, Tensor]:
        """Recommend from a continuous search space without cardinality constraints.

        Args:
            subspace_continuous: The continuous subspace from which to generate
                recommendations.
            batch_size: The size of the recommendation batch.

        Returns:
            The recommendations and corresponding acquisition values.

        Raises:
            ValueError: If the continuous search space has cardinality constraints.
        """
        import torch
        from botorch.optim import optimize_acqf

        if subspace_continuous.constraints_cardinality:
            raise ValueError(
                f"'{self._recommend_continuous_without_cardinality_constraints.__name__}' "  # noqa: E501
                f"expects a subspace without constraints of type "
                f"'{ContinuousCardinalityConstraint.__name__}'. "
            )

        fixed_parameters = {
            idx: p.value
            for (idx, p) in enumerate(subspace_continuous.parameters)
            if isinstance(p, _FixedNumericalContinuousParameter)
        }

        # NOTE: The explicit `or None` conversion is added as an additional safety net
        #   because it is unclear if the corresponding presence checks for these
        #   arguments is correctly implemented in all invoked BoTorch subroutines.
        #   For details: https://github.com/pytorch/botorch/issues/2042
        points, acqf_values = optimize_acqf(
            acq_function=self._botorch_acqf,
            bounds=torch.from_numpy(subspace_continuous.comp_rep_bounds.values),
            q=batch_size,
            num_restarts=self.n_restarts,
            raw_samples=self.n_raw_samples,
            fixed_features=fixed_parameters or None,
            equality_constraints=[
                c.to_botorch(subspace_continuous.parameters)
                for c in subspace_continuous.constraints_lin_eq
            ]
            or None,
            inequality_constraints=[
                c.to_botorch(subspace_continuous.parameters)
                for c in subspace_continuous.constraints_lin_ineq
            ]
            or None,
            sequential=self.sequential_continuous,
        )
        return points, acqf_values

    @override
    def _recommend_hybrid(
        self,
        searchspace: SearchSpace,
        candidates_exp: pd.DataFrame,
        batch_size: int,
    ) -> pd.DataFrame:
        """Recommend points using the ``optimize_acqf_mixed`` function of BoTorch.

        This functions samples points from the discrete subspace, performs optimization
        in the continuous subspace with these points being fixed and returns the best
        found solution.

        **Important**: This performs a brute-force calculation by fixing every possible
        assignment of discrete variables and optimizing the continuous subspace for
        each of them. It is thus computationally expensive.

        **Note**: This function implicitly assumes that discrete search space parts in
        the respective data frame come first and continuous parts come second.

        Args:
            searchspace: The search space in which the recommendations should be made.
            candidates_exp: The experimental representation of the candidates
                of the discrete subspace.
            batch_size: The size of the calculated batch.

        Raises:
            IncompatibleAcquisitionFunctionError: If a non-Monte Carlo acquisition
                function is used with a batch size > 1.

        Returns:
            The recommended points.
        """
        assert self._objective is not None
        if (
            batch_size > 1
            and not self._get_acquisition_function(self._objective).supports_batching
        ):
            raise IncompatibleAcquisitionFunctionError(
                f"The '{self.__class__.__name__}' only works with Monte Carlo "
                f"acquisition functions for batch sizes > 1."
            )

        import torch
        from botorch.optim import optimize_acqf_mixed

        # Transform discrete candidates
        candidates_comp = searchspace.discrete.transform(candidates_exp)

        # Calculate the number of samples from the given percentage
        n_candidates = math.ceil(self.sampling_percentage * len(candidates_comp.index))

        # Potential sampling of discrete candidates
        if self.hybrid_sampler is not None:
            candidates_comp = sample_numerical_df(
                candidates_comp, n_candidates, method=self.hybrid_sampler
            )

        # Prepare all considered discrete configurations in the
        # List[Dict[int, float]] format expected by BoTorch.
        num_comp_columns = len(candidates_comp.columns)
        candidates_comp.columns = list(range(num_comp_columns))  # type: ignore
        fixed_features_list = candidates_comp.to_dict("records")

        # Actual call of the BoTorch optimization routine
        # NOTE: The explicit `or None` conversion is added as an additional safety net
        #   because it is unclear if the corresponding presence checks for these
        #   arguments is correctly implemented in all invoked BoTorch subroutines.
        #   For details: https://github.com/pytorch/botorch/issues/2042
        points, _ = optimize_acqf_mixed(
            acq_function=self._botorch_acqf,
            bounds=torch.from_numpy(searchspace.comp_rep_bounds.values),
            q=batch_size,
            num_restarts=self.n_restarts,
            raw_samples=self.n_raw_samples,
            fixed_features_list=fixed_features_list,
            equality_constraints=[
                c.to_botorch(
                    searchspace.continuous.parameters,
                    idx_offset=len(candidates_comp.columns),
                )
                for c in searchspace.continuous.constraints_lin_eq
            ]
            or None,
            inequality_constraints=[
                c.to_botorch(
                    searchspace.continuous.parameters,
                    idx_offset=num_comp_columns,
                )
                for c in searchspace.continuous.constraints_lin_ineq
            ]
            or None,
        )

        # Align candidates with search space index. Done via including the search space
        # index during the merge, which is used later for back-translation into the
        # experimental representation
        merged = pd.merge(
            pd.DataFrame(points),
            candidates_comp.reset_index(),
            on=list(candidates_comp.columns),
            how="left",
        ).set_index("index")

        # Get experimental representation of discrete part
        rec_disc_exp = searchspace.discrete.exp_rep.loc[merged.index]

        # Combine discrete and continuous parts
        rec_exp = pd.concat(
            [
                rec_disc_exp,
                merged.iloc[:, num_comp_columns:].set_axis(
                    searchspace.continuous.parameter_names, axis=1
                ),
            ],
            axis=1,
        )

        return rec_exp

    def _optimize_continuous_subspaces(
        self, subspaces: Iterable[SubspaceContinuous], batch_size: int
    ) -> tuple[Tensor, Tensor]:
        """Find the optimum candidates from multiple continuous subspaces.

        Important:
            Subspaces without feasible solutions will be silently ignored. If none of
            the subspaces has a feasible solution, an exception will be raised.

        Args:
            subspaces: The subspaces to consider for the optimization.
            batch_size: The number of points to be recommended.

        Raises:
            InfeasibilityError: If none of the subspaces has a feasible solution.

        Returns:
            The batch of candidates and the corresponding acquisition value.
        """
        import torch
        from botorch.exceptions.errors import InfeasibilityError as BoInfeasibilityError

        acqf_values_all: list[Tensor] = []
        points_all: list[Tensor] = []

        for subspace in subspaces:
            try:
                # Optimize the acquisition function
                # Note: We explicitly evaluate the acqf function for the batch because
                #   the object returned by the optimization routine may contain joint or
                #   individual acquisition values, depending on the whether sequential
                #   or joint optimization is applied
                p, _ = self._recommend_continuous_torch(subspace, batch_size)
                with torch.no_grad():
                    acqf = self._botorch_acqf(p)

                # Append optimization results
                points_all.append(p)
                acqf_values_all.append(acqf)

            # The optimization problem may be infeasible in certain subspaces
            except BoInfeasibilityError:
                pass

        if not points_all:
            raise InfeasibilityError(
                "No feasible solution could be found. Potentially the specified "
                "constraints are too restrictive, i.e. there may be too many "
                "constraints or thresholds may have been set too tightly. "
                "Considered relaxing the constraints to improve the chances "
                "of finding a feasible solution."
            )

        # Find the best option f
        best_idx = np.argmax(acqf_values_all)
        points = points_all[best_idx]
        acqf_value = acqf_values_all[best_idx]

        return points, acqf_value


# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Deprecation tests."""

import warnings
from unittest.mock import patch

import pandas as pd
import pytest
from pytest import param

from baybe._optional.info import CHEM_INSTALLED
from baybe.constraints import (
    ContinuousLinearConstraint,
    ContinuousLinearEqualityConstraint,
    ContinuousLinearInequalityConstraint,
)
from baybe.constraints.base import Constraint
from baybe.exceptions import DeprecationError
from baybe.objectives.desirability import DesirabilityObjective
from baybe.objectives.single import SingleTargetObjective
from baybe.parameters.enum import SubstanceEncoding
from baybe.parameters.numerical import (
    NumericalContinuousParameter,
    NumericalDiscreteParameter,
)
from baybe.recommenders.meta.sequential import TwoPhaseMetaRecommender
from baybe.recommenders.pure.bayesian import (
    BotorchRecommender,
    SequentialGreedyRecommender,
)
from baybe.recommenders.pure.nonpredictive.sampling import RandomRecommender
from baybe.searchspace.continuous import SubspaceContinuous
from baybe.searchspace.discrete import SubspaceDiscrete
from baybe.searchspace.validation import get_transform_parameters
from baybe.targets.binary import BinaryTarget
from baybe.targets.numerical import NumericalTarget


def test_sequentialgreedyrecommender_class():
    """Using the deprecated `SequentialGreedyRecommender` class raises a warning."""
    with pytest.warns(DeprecationWarning):
        SequentialGreedyRecommender()


def test_samples_random():
    """Using the deprecated `samples_random` method raises a warning."""
    with pytest.warns(DeprecationWarning):
        parameters = [NumericalContinuousParameter("x", (0, 1))]
        SubspaceContinuous(parameters).samples_random(n_points=1)


def test_samples_full_factorial():
    """Using the deprecated `samples_full_factorial` method raises a warning."""
    with pytest.warns(DeprecationWarning):
        parameters = [NumericalContinuousParameter("x", (0, 1))]
        SubspaceContinuous(parameters).samples_full_factorial(n_points=1)


def test_subspace_transform_interface(searchspace):
    """Using the deprecated transform interface raises a warning."""
    # Not providing `allow_extra` when there are additional columns
    with pytest.warns(DeprecationWarning):
        searchspace.discrete.transform(
            pd.DataFrame(columns=["additional", *searchspace.discrete.exp_rep.columns]),
        )
    with pytest.warns(DeprecationWarning):
        searchspace.continuous.transform(
            pd.DataFrame(columns=["additional", *searchspace.discrete.exp_rep.columns]),
        )

    # Passing dataframe via `data`
    with pytest.warns(DeprecationWarning):
        searchspace.discrete.transform(
            data=searchspace.discrete.exp_rep, allow_extra=True
        )
    with pytest.warns(DeprecationWarning):
        searchspace.continuous.transform(
            data=searchspace.discrete.exp_rep, allow_extra=True
        )


def test_surrogate_registration():
    """Using the deprecated registration mechanism raises a warning."""
    from baybe.surrogates import register_custom_architecture

    with pytest.raises(DeprecationError):
        register_custom_architecture()


def test_surrogate_access():
    """Public attribute access to the surrogate model raises a warning."""
    recommender = BotorchRecommender()
    with pytest.warns(DeprecationWarning):
        recommender.surrogate_model


def test_continuous_linear_eq_constraint():
    """Usage of deprecated continuous linear eq constraint raises a warning."""
    with pytest.warns(DeprecationWarning):
        ContinuousLinearEqualityConstraint(["p1", "p2"])


def test_continuous_linear_inq_constraint():
    """Usage of deprecated continuous linear ineq constraint raises a warning."""
    with pytest.warns(DeprecationWarning):
        ContinuousLinearInequalityConstraint(["p1", "p2"])


@pytest.mark.parametrize(
    ("type_", "op"),
    [
        ("ContinuousLinearEqualityConstraint", "="),
        ("ContinuousLinearInequalityConstraint", ">="),
    ],
    ids=["lin_eq", "lin_ineq"],
)
def test_constraint_config_deserialization(type_, op):
    """The deprecated constraint config format can still be parsed."""
    config = """
    {
        "type": "__replace__",
        "parameters": ["p1", "p2", "p3"],
        "coefficients": [1.0, 2.0, 3.0],
        "rhs": 2.0
    }
    """
    config = config.replace("__replace__", type_)

    expected = ContinuousLinearConstraint(
        parameters=["p1", "p2", "p3"],
        operator=op,
        coefficients=[1.0, 2.0, 3.0],
        rhs=2.0,
    )

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        actual = Constraint.from_json(config)
    assert expected == actual, (expected, actual)


def test_objective_transform_interface():
    """Using the deprecated transform interface raises a warning."""
    single = SingleTargetObjective(NumericalTarget("A", "MAX"))
    desirability = DesirabilityObjective(
        [
            NumericalTarget("A", "MAX", (0, 1)),
            NumericalTarget("B", "MIN", (-1, 1)),
        ]
    )

    # Not providing `allow_extra` when there are additional columns
    with pytest.warns(DeprecationWarning):
        single.transform(
            pd.DataFrame(columns=["A", "additional"]),
        )
    with pytest.warns(DeprecationWarning):
        desirability.transform(
            pd.DataFrame(columns=["A", "B", "additional"]),
        )

    # Passing dataframe via `data`
    with pytest.warns(DeprecationWarning):
        single.transform(data=pd.DataFrame(columns=["A"]), allow_extra=True)
    with pytest.warns(DeprecationWarning):
        desirability.transform(data=pd.DataFrame(columns=["A", "B"]), allow_extra=True)


def test_deprecated_get_transform_parameters():
    """Using the deprecated utility raises a warning."""
    with pytest.warns(
        DeprecationWarning, match="'get_transform_parameters' has been deprecated"
    ):
        get_transform_parameters(pd.DataFrame(), [])


def test_target_transform_interface():
    """Using the deprecated transform interface raises a warning."""
    numerical = NumericalTarget("num", "MAX")
    binary = BinaryTarget("bin")

    # Passing dataframe via `data`
    with pytest.warns(DeprecationWarning):
        numerical.transform(data=pd.DataFrame(columns=["num"]))
    with pytest.warns(DeprecationWarning):
        binary.transform(data=pd.DataFrame(columns=["bin"]))


@pytest.mark.parametrize(
    ("deprecated", "replacement"),
    [
        param(SubstanceEncoding.MORGAN_FP, "ECFPFingerprint", id="morgan"),
        param(SubstanceEncoding.RDKIT, "RDKit2DDescriptorsFingerprint", id="rdkit"),
    ],
)
@pytest.mark.skipif(
    not CHEM_INSTALLED, reason="Optional chem dependency not installed."
)
def test_deprecated_encodings(deprecated, replacement):
    """Deprecated encoding raises a warning and uses correct replacement."""
    import skfp.fingerprints

    from baybe.utils.chemistry import smiles_to_fingerprint_features

    path = f"skfp.fingerprints.{replacement}"

    with patch(path, wraps=getattr(skfp.fingerprints, replacement)) as patched:
        # Assert warning
        with pytest.warns(DeprecationWarning):
            smiles_to_fingerprint_features(["C"], deprecated)

        # Check that equivalent is used instead of deprecated encoding
        if deprecated is SubstanceEncoding.MORGAN_FP:
            patched.assert_called_once_with(**{"fp_size": 1024, "radius": 4})
        else:
            patched.assert_called_once()


def test_migrated_metadata_attribute():
    """Accessing the migrated metadata search space attribute raises an error."""
    with pytest.raises(DeprecationError, match="no longer carry any metadata"):
        subspace = SubspaceDiscrete.from_parameter(
            NumericalDiscreteParameter("p", [0, 1])
        )
        subspace.metadata


def test_deprecated_meta_recommender_methods():
    """Calling the deprecated methods of meta recommender raises an error."""
    recommender = TwoPhaseMetaRecommender()
    with pytest.raises(DeprecationError, match="has been deprecated."):
        recommender.get_current_recommender()
    with pytest.raises(DeprecationError, match="has been deprecated"):
        recommender.get_next_recommender()


@pytest.mark.parametrize(
    "flag",
    [
        "allow_repeated_recommendations",
        "allow_recommending_already_measured",
        "allow_recommending_pending_experiments",
    ],
)
@pytest.mark.parametrize("recommender_cls", [RandomRecommender, BotorchRecommender])
def test_migrated_allow_flags(flag, recommender_cls):
    """Passing and accessing the migrated 'allow_*' flags raises an error."""
    with pytest.raises(DeprecationError, match=r"Passing 'allow_\*' flags"):
        recommender_cls(**{flag: True})

    with pytest.raises(DeprecationError, match=f"The attribute '{flag}' is no longer"):
        getattr(recommender_cls(), flag)



"""Classes for persisting benchmark results."""

from __future__ import annotations

import json
import os
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Protocol

import boto3
import boto3.session
from attr import define, field
from attrs.validators import instance_of
from boto3.session import Session
from typing_extensions import override

from benchmarks import Result

VARNAME_BENCHMARKING_PERSISTENCE_PATH = "BAYBE_BENCHMARKING_PERSISTENCE_PATH"


class PathStrategy(Enum):
    """Specifies the way a file path is constructed."""

    HIERARCHICAL = "HIERARCHICAL"
    """Hierarchical path construction using folders."""

    FLAT = "FLAT"
    """Flat path construction using a file name only."""


@define
class PathConstructor:
    """A class to construct the file path of a result object.

    The class encapsulates the construction of a file path depending
    on where the object should be stored. Since different storage backends have
    different requirements for the path, and some like the used S3 bucket
    do not support folders, this class uses its variables
    to construct a `Path` object depending on the strategy.

    For compatibility reasons, the path is sanitized to contain only
    lower and upper case letters, digits as well as the symbols '.' and '-'"
    """

    benchmark_name: str = field(validator=instance_of(str))
    """The name of the benchmark for which the path should be constructed."""

    branch: str = field(
        converter=lambda x: x or "-branchless-",
        validator=instance_of(str),
    )
    """The branch checked out at benchmark execution time.
    In case of detached head state the branch is set to '-branchless-'."""

    latest_baybe_tag: str = field(validator=instance_of(str))
    """The latest BayBE version tag existing at benchmark execution time."""

    execution_date_time: datetime = field(validator=instance_of(datetime))
    """The date and time when the benchmark was executed."""

    commit_hash: str = field(validator=instance_of(str))
    """The hash of the commit checked out at benchmark execution time."""

    def _sanitize_string(self, string: str) -> str:
        """Replace disallowed characters for filesystems in the given string."""
        ALLOWED_CHARS = (
            "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-."
        )
        return "".join([char if char in ALLOWED_CHARS else "-" for char in string])

    @classmethod
    def from_result(cls, result: Result) -> PathConstructor:
        """Create a path constructor from result.

        Args:
            result: The result of the benchmark.

        Returns:
            The path constructor.
        """
        benchmark_name = result.benchmark_identifier
        start_datetime = result.metadata.start_datetime
        commit_hash = result.metadata.commit_hash
        latest_baybe_tag = result.metadata.latest_baybe_tag
        branch = result.metadata.branch

        return PathConstructor(
            benchmark_name=benchmark_name,
            latest_baybe_tag=latest_baybe_tag,
            commit_hash=commit_hash,
            execution_date_time=start_datetime,
            branch=branch,
        )

    def get_path(self, strategy: PathStrategy) -> Path:
        """Construct the path of a result object.

        Creates the path depending on the chosen strategy by concatenating
        the experiment identifier, the branch, the BayBE version,
        the start date (without time information) and the commit hash.

        Args:
            strategy: The strategy to construct the path.

        Returns:
            The path to persist the object. Can be a file name or
            a folder path as a string.
        """
        separator = "/" if strategy is PathStrategy.HIERARCHICAL else "_"

        file_usable_date = self.execution_date_time.strftime("%Y-%m-%d")

        components = [
            self.benchmark_name,
            self.branch,
            self.latest_baybe_tag,
            file_usable_date,
            self.commit_hash,
        ]

        sanitized_components = [
            self._sanitize_string(component) for component in components
        ]
        path = separator.join(sanitized_components) + separator + "result.json"

        return Path(path)


class ObjectStorage(Protocol):
    """Interface for interacting with storage."""

    def write_json(self, object: dict, path_constructor: PathConstructor) -> None:
        """Store a JSON serializable dictionary according to the path_constructor.

        If the respective object exists, it will be overwritten.

        Args:
            object: The object to be persisted.
            path_constructor: The path constructor creating the path for the object.
        """


@define
class S3ObjectStorage(ObjectStorage):
    """Class for persisting objects in an S3 bucket."""

    _bucket_name: str = field(validator=instance_of(str), init=False)
    """The name of the S3 bucket where the objects are stored."""

    _object_session: Session = field(factory=boto3.session.Session)
    """The boto3 session object. Loads the required credentials
    from environment variables."""

    @_bucket_name.default
    def _default_bucket_name(self) -> str:
        """Get the bucket name from the environment variables."""
        if VARNAME_BENCHMARKING_PERSISTENCE_PATH not in os.environ:
            raise ValueError(
                f"No S3 bucket name provided. Please provide the "
                f"bucket name by setting the environment variable "
                f"'{VARNAME_BENCHMARKING_PERSISTENCE_PATH}'."
            )
        return os.environ[VARNAME_BENCHMARKING_PERSISTENCE_PATH]

    @override
    def write_json(self, object: dict, path_constructor: PathConstructor) -> None:
        """Store a JSON serializable dictionary in an S3 bucket.

        The S3-key of the JSON is created from
        the path_constructor. If the key already exists, it will be overwritten.

        Args:
            object: The object to be persisted.
            path_constructor: The path constructor creating the path for the object.
        """
        client = self._object_session.client("s3")

        key = path_constructor.get_path(strategy=PathStrategy.HIERARCHICAL)

        client.put_object(
            Bucket=self._bucket_name,
            Key=key.as_posix(),
            Body=json.dumps(object),
            ContentType="application/json",
        )


@define
class LocalFileObjectStorage(ObjectStorage):
    """Class for persisting objects locally."""

    folder_path_prefix: Path = field(converter=Path, default=Path("."))
    """The prefix of the folder path where the results are stored."""

    @folder_path_prefix.validator
    def _folder_path_prefix_validator(self, _, folder_path_prefix: Path) -> None:
        """Validate the existence of the path."""
        if not folder_path_prefix.exists():
            raise FileNotFoundError(
                f"The folder path '{folder_path_prefix.resolve()}' does not exist."
            )

    @override
    def write_json(self, object: dict, path_constructor: PathConstructor) -> None:
        """Store a JSON serializable dictionary in the local file system.

        If the respective file exists, it will be overwritten.

        Args:
            object: The object to be persisted.
            path_constructor: The path constructor creating the path for the object.
        """
        path_object = self.folder_path_prefix.joinpath(
            path_constructor.get_path(strategy=PathStrategy.FLAT)
        )
        with open(path_object.resolve(), "w") as file:
            json.dump(object, file)



"""Benchmark on aryl halides data with two distinct aryl halides as TL tasks.

This module provides the basic structure for creating different
benchmarks by changing the source and target tasks. The benchmark compares TL and
non-TL campaigns.

By convention, the benchmarks name use the format
"sou_<Source tasks>_tar_<Target task>".
"""

from __future__ import annotations

from collections.abc import Sequence

import pandas as pd

from baybe.campaign import Campaign
from baybe.objectives import SingleTargetObjective
from baybe.parameters import SubstanceParameter, TaskParameter
from baybe.parameters.base import DiscreteParameter
from baybe.searchspace import SearchSpace
from baybe.simulation import simulate_scenarios
from baybe.targets import NumericalTarget
from benchmarks.data.utils import DATA_PATH
from benchmarks.definition import ConvergenceBenchmarkSettings


def load_data() -> pd.DataFrame:
    """Load the data for the benchmark."""
    relevant_columns = [
        "base",
        "ligand",
        "additive",
        "ligand_smiles",
        "base_smiles",
        "additive_smiles",
        "aryl_halide",
        "yield",
    ]
    data = pd.read_table(
        DATA_PATH / "aryl_halide" / "data.csv", sep=",", usecols=relevant_columns
    ).dropna(subset=["base", "ligand", "additive", "aryl_halide"])
    return data


def make_searchspace(
    data: pd.DataFrame,
    target_tasks: Sequence[str] | None = None,
    source_tasks: Sequence[str] | None = None,
) -> SearchSpace:
    """Create the search space for the benchmark."""
    params: list[DiscreteParameter] = [
        SubstanceParameter(
            name=substance,
            data=dict(zip(data[substance], data[f"{substance}_smiles"])),
            encoding="MORDRED",
        )
        for substance in ["base", "ligand", "additive"]
    ]
    if target_tasks is not None and source_tasks is not None:
        all_tasks = [*source_tasks, *target_tasks]
        all_tasks = [*source_tasks, *target_tasks]
        params.append(
            TaskParameter(
                name="aryl_halide",
                values=all_tasks,
                active_values=target_tasks,
            )
        )
    return SearchSpace.from_product(parameters=params)


def make_lookup(data: pd.DataFrame, target_tasks: Sequence[str]) -> pd.DataFrame:
    """Create the lookup for the benchmark.

    Without the filtering, there would be multiple entries for the same parameter
    configuration. Since this might yield issues for the non-transfer learning
    campaigns, we filter the data to only include the target tasks.
    """
    return data[data["aryl_halide"].isin(target_tasks)]


def make_initial_data(data: pd.DataFrame, source_tasks: Sequence[str]) -> pd.DataFrame:
    """Create the initial data for the benchmark."""
    return data[data["aryl_halide"].isin(source_tasks)]


def aryl_halide_tl_substance_benchmark(
    settings: ConvergenceBenchmarkSettings,
    source_tasks: Sequence[str],
    target_tasks: Sequence[str],
    percentages: Sequence[float],
) -> pd.DataFrame:
    """Abstract benchmark function comparing TL and non-TL campaigns.

    Inputs:
        base:           Substance parameter
        ligand:         Substance parameter
        additive:       Substance parameter
        aryl_halide:    Task parameter
    Output:             Continuous (yield)
    Objective:          Maximization
    """
    data = load_data()

    searchspace = make_searchspace(
        data=data,
        source_tasks=source_tasks,
        target_tasks=target_tasks,
    )
    searchspace_nontl = make_searchspace(data=data)

    lookup = make_lookup(data, target_tasks)
    initial_data = make_initial_data(data, source_tasks)
    objective = SingleTargetObjective(NumericalTarget(name="yield", mode="MAX"))

    tl_campaign = Campaign(
        searchspace=searchspace,
        objective=objective,
    )
    nontl_campaign = Campaign(searchspace=searchspace_nontl, objective=objective)

    results = []
    for p in percentages:
        results.append(
            simulate_scenarios(
                {
                    f"{int(100 * p)}": tl_campaign,
                    f"{int(100 * p)}_naive": nontl_campaign,
                },
                lookup,
                initial_data=[
                    initial_data.sample(frac=p) for _ in range(settings.n_mc_iterations)
                ],
                batch_size=settings.batch_size,
                n_doe_iterations=settings.n_doe_iterations,
                impute_mode="error",
            )
        )
    results.append(
        simulate_scenarios(
            {"0": tl_campaign, "0_naive": nontl_campaign},
            lookup,
            batch_size=settings.batch_size,
            n_doe_iterations=settings.n_doe_iterations,
            n_mc_iterations=settings.n_mc_iterations,
            impute_mode="error",
        )
    )
    return pd.concat(results)



"""Functionality for building custom surrogates.

Note that ONNX surrogate models cannot be retrained. However, having the surrogates
raise a ``NotImplementedError`` would currently break the code since
:class:`baybe.recommenders.pure.bayesian.base.BayesianRecommender` assumes that
surrogates can be trained and attempts to do so for each new DOE iteration.

It is planned to solve this issue in the future.
"""

from __future__ import annotations

import gc
from typing import TYPE_CHECKING, ClassVar, NoReturn

from attrs import define, field, validators
from typing_extensions import override

from baybe.exceptions import DeprecationError
from baybe.parameters import (
    CategoricalEncoding,
    CategoricalParameter,
    CustomDiscreteParameter,
    NumericalContinuousParameter,
    NumericalDiscreteParameter,
    TaskParameter,
)
from baybe.searchspace import SearchSpace
from baybe.surrogates.base import IndependentGaussianSurrogate
from baybe.surrogates.utils import batchify_mean_var_prediction
from baybe.utils.conversion import to_string
from baybe.utils.numerical import DTypeFloatONNX

if TYPE_CHECKING:
    from torch import Tensor


def register_custom_architecture(*args, **kwargs) -> NoReturn:
    """Deprecated! Raises an error when used."""  # noqa: D401
    raise DeprecationError(
        "The 'register_custom_architecture' decorator is no longer available. "
        "Use :class:`baybe.surrogates.base.SurrogateProtocol` instead to define "
        "your custom architectures."
    )


@define(kw_only=True)
class CustomONNXSurrogate(IndependentGaussianSurrogate):
    """A wrapper class for custom pretrained surrogate models.

    Note that these surrogates cannot be retrained.
    """

    supports_transfer_learning: ClassVar[bool] = False
    # See base class.

    onnx_input_name: str = field(validator=validators.instance_of(str))
    """The input name used for constructing the ONNX str."""

    onnx_str: bytes = field(validator=validators.instance_of(bytes))
    """The ONNX byte str representing the model."""

    # TODO: type should be `onnxruntime.InferenceSession` but is currently
    #   omitted due to: https://github.com/python-attrs/cattrs/issues/531
    _model = field(init=False, default=None, eq=False)
    """The actual model."""

    @override
    @batchify_mean_var_prediction
    def _estimate_moments(
        self, candidates_comp_scaled: Tensor, /
    ) -> tuple[Tensor, Tensor]:
        import torch

        from baybe._optional.onnx import onnxruntime as ort
        from baybe.utils.torch import DTypeFloatTorch

        model_inputs = {
            self.onnx_input_name: candidates_comp_scaled.numpy().astype(DTypeFloatONNX)
        }
        if self._model is None:
            self._model = ort.InferenceSession(self.onnx_str)
        results = self._model.run(None, model_inputs)

        # IMPROVE: At the moment, we assume that the second model output contains
        #   standard deviations. Currently, most available ONNX converters care
        #   about the mean only and it's not clear how this will be handled in the
        #   future. Once there are more choices available, this should be revisited.
        return (
            torch.from_numpy(results[0]).to(DTypeFloatTorch),
            torch.from_numpy(results[1]).pow(2).to(DTypeFloatTorch),
        )

    @override
    def _fit(self, train_x: Tensor, train_y: Tensor) -> None:
        # TODO: This method actually needs to raise a NotImplementedError because
        #   ONNX surrogate models cannot be retrained. However, this would currently
        #   break the code since `BayesianRecommender` assumes that surrogates
        #   can be trained and attempts to do so for each new DOE iteration.
        #   Therefore, a refactoring is required in order to properly incorporate
        #   "static" surrogates and account for them in the exposed APIs.
        pass

    @classmethod
    def validate_compatibility(cls, searchspace: SearchSpace) -> None:
        """Validate if the class is compatible with a given search space.

        Args:
            searchspace: The search space to be tested for compatibility.

        Raises:
            TypeError: If the search space is incompatible with the class.
        """
        if not all(
            isinstance(
                p,
                (
                    NumericalContinuousParameter,
                    NumericalDiscreteParameter,
                    TaskParameter,
                ),
            )
            or (isinstance(p, CustomDiscreteParameter) and not p.decorrelate)
            or (
                isinstance(p, CategoricalParameter)
                and p.encoding is CategoricalEncoding.INT
            )
            for p in searchspace.parameters
        ):
            raise TypeError(
                f"To prevent potential hard-to-detect bugs that stem from wrong "
                f"wiring of model inputs, {cls.__name__} "
                f"is currently restricted for use with parameters that have "
                f"a one-dimensional computational representation or "
                f"{CustomDiscreteParameter.__name__}."
            )

    @override
    def __str__(self) -> str:
        fields = [to_string("ONNX input name", self.onnx_input_name, single_line=True)]
        return to_string(super().__str__(), *fields)


# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Test serialization of surrogates."""

import pytest

from baybe._optional.info import NGBOOST_INSTALLED, ONNX_INSTALLED
from baybe.surrogates.base import Surrogate
from baybe.surrogates.composite import CompositeSurrogate
from baybe.surrogates.custom import CustomONNXSurrogate
from baybe.surrogates.gaussian_process.core import GaussianProcessSurrogate
from baybe.surrogates.ngboost import NGBoostSurrogate
from baybe.surrogates.random_forest import RandomForestSurrogate
from baybe.utils.basic import get_subclasses


@pytest.mark.parametrize("surrogate_cls", get_subclasses(Surrogate))
def test_surrogate_serialization(request, surrogate_cls):
    """A serialization roundtrip yields an equivalent object."""
    if issubclass(surrogate_cls, CustomONNXSurrogate):
        if not ONNX_INSTALLED:
            pytest.skip("Optional onnx dependency not installed.")
        surrogate = request.getfixturevalue("onnx_surrogate")
    elif issubclass(surrogate_cls, NGBoostSurrogate) and not NGBOOST_INSTALLED:
        pytest.skip("Optional ngboost dependency not installed.")
    else:
        surrogate = surrogate_cls()

    string = surrogate.to_json()
    surrogate2 = Surrogate.from_json(string)
    assert surrogate == surrogate2, (surrogate, surrogate2)


@pytest.mark.parametrize(
    "surrogate",
    [
        CompositeSurrogate(
            {"A": RandomForestSurrogate(), "B": GaussianProcessSurrogate()}
        ),
        CompositeSurrogate.from_replication(GaussianProcessSurrogate()),
    ],
    ids=["via_init", "via_template"],
)
def test_composite_surrogate_serialization(surrogate):
    """A serialization roundtrip yields an equivalent object."""
    string = surrogate.to_json()
    surrogate2 = CompositeSurrogate.from_json(string)
    assert surrogate == surrogate2, (surrogate, surrogate2)



"""Hypothesis strategies for targets."""

import hypothesis.strategies as st

from baybe.targets.binary import BinaryTarget
from baybe.targets.enum import TargetMode
from baybe.targets.numerical import _VALID_TRANSFORMATIONS, NumericalTarget
from baybe.utils.interval import Interval

from .utils import intervals as st_intervals

target_name = st.text(min_size=1)
"""A strategy that generates target names."""


@st.composite
def numerical_targets(
    draw: st.DrawFn, bounds_strategy: st.SearchStrategy[Interval] | None = None
):
    """Generate :class:`baybe.targets.numerical.NumericalTarget`.

    Args:
        draw: Hypothesis draw object.
        bounds_strategy: An optional strategy for generating the target bounds.

    Returns:
        _type_: _description_
    """
    name = draw(target_name)
    mode = draw(st.sampled_from(TargetMode))
    transformation = draw(st.sampled_from(_VALID_TRANSFORMATIONS[mode]))
    if bounds_strategy is None:
        bounds_strategy = st_intervals(
            exclude_half_bounded=True,
            exclude_fully_unbounded=mode is TargetMode.MATCH
            or transformation is not None,
        )
    bounds = draw(bounds_strategy)

    return NumericalTarget(
        name=name, mode=mode, bounds=bounds, transformation=transformation
    )


choice_values = st.one_of(
    [st.booleans(), st.integers(), st.floats(allow_nan=False), st.text()]
)
"""A strategy that generates choice values."""


@st.composite
def binary_targets(draw: st.DrawFn):
    """A strategy that generates binary targets."""
    name = draw(target_name)
    choices = draw(st.lists(choice_values, min_size=2, max_size=2, unique=True))
    return BinaryTarget(name, success_value=choices[0], failure_value=choices[1])


targets = st.one_of([binary_targets(), numerical_targets()])
"""A strategy that generates targets."""
